(function capsule(window) {
  function defineLibrary() {
    const Vec3 = VemaLib.Vec3;
    const Matrix4 = VemaLib.Matrix4;
    const BoidsLib = {};

    class Boids {
      constructor(N, type) {
        this.width = 5.0;
        this.height = 8;

        this.offsetx = 0.0;
        this.offsety = 1.5;
        this.offsetz = 0.0;
        this.maxHeight = 100;
        this.minHeight = -10;
        this.center = new Vec3(this.offsetx, this.offsety, this.offsetz);
        this.noh = N;
        this.maxN = N;
        this.NB = 1;
        this.N = 0;
        this.vel = [];
        this.pos = [];
        this.color = [];
        this.size = [];
        this.clonelist = [];
        this.types = [];
        this.type = type;

        this.markForDelete = null;

        this.rule1 = true;
        this.rule2 = true;
        this.rule3 = true;
        this.rule4 = true;

        this.rule5 = true;

        this.rule6 = true;
        this.rule10 = true;
        this.rule11 = true;

        this.rule12 = false;
        this.rule13 = true;
        this.rule7 = true;
        this.rule8 = true;
        this.rule9 = true;

    // min dist of approach
        this.minDist = 0.4;
    // max magnitude of velocities calculated by "rules"
        this.maxRuleVel = 0.0005;
    // max maginitude of final velocity
        this.maxVel = 0.02;
    /*
    if(this.type==1){
      this.minDist =0.02;
      // max magnitude of velocities calculated by "rules"
      this.maxRuleVel = 0.0001;
      // max maginitude of final velocity
      this.maxVel = 0.002;
    }
    */
      }

      cloneBoid(i) {
    // console.log("clone "+i);
        this.clonelist.push(i);
      }

      birthBoids() {
        if (this.clonelist.length > 0) {
          for (let j = 0; j < this.clonelist.length; j++) {
        // console.log("plop");
            const i = this.clonelist[j];
        // console.log("new "+i);
            this.vel.push(this.vel[i].multiple(-1.0));
            this.pos.push(this.pos[i].clone().subtract(this.vel[i]));
            this.color.push(this.color[i].clone());
            this.types.push(this.types[i]);
            this.size.push(this.size[i]);
            this.N += 1;
          }
          this.clonelist = [];
        }
      }

      addBoids(x, y, z) {
    // console.log("p "+p+" r "+r);
        if (this.N < this.maxN) {
      // console.log("test"+(ground(x,y)+1.2)*0.5);
      // console.log("add boid");
          const angle = 2.0 * Math.PI * Math.random();

          const anglesup = 0.2 * Math.PI * Math.random();

          this.vel.push(new Vec3(Math.sin(angle), Math.sin(anglesup), Math.cos(angle)));
          this.pos.push(new Vec3(x, y, z));
          const r = Math.random() * 0.1;

          this.color.push(new Vec3(0.1 + r + Math.random() * 0.1, 0.1 + r + Math.random() * 0.1, 0.1 + r + Math.random() * 0.1));

          this.types.push(this.type);
          this.size.push(0.15 + 0.05 * Math.random());
          this.N += 1;
        }
      }


      limitVec(vec, maxVal) {
    // """limit magnitide of 2D vector"""
        const mag = vec.length;
        if (mag > maxVal) {
          vec.f[0] = vec.f[0] * maxVal / mag;
          vec.f[1] = vec.f[1] * maxVal / mag;
          vec.f[2] = vec.f[2] * maxVal / mag;
        }
      }
      limit(X, maxVal) {
    // """limit magnitide of 2D vectors in array X to maxValue"""
        for (let i = 0; i < this.N; i++) {
          this.limitVec(X[i], maxVal);
        }
      }

      reaper() {
        if (this.markForDelete !== null) {
          for (let i = this.markForDelete.length - 1; i >= 0; i--) { this.removeBoid(this.markForDelete[i]); }
          this.markForDelete = null;
        }
      }


      removeBoid(selectedindex) {
        this.pos.splice(selectedindex, 1);
        this.vel.splice(selectedindex, 1);
        this.color.splice(selectedindex, 1);
        this.size.splice(selectedindex, 1);
        this.types.splice(selectedindex, 1);
        this.N -= 1;
      }
      calculateDistanceMatrix(pos) {
        const m = [];
        for (let i = 0; i < this.N; i++) {
          const r = [];
          for (let j = 0; j < this.N; j++) {
            r.push(0.0);
          }
          m.push(r);
        }
        for (let i = 0; i < this.N; i++) {
          for (let j = i + 1; j < this.N; j++) {
            const l = (pos[i].subtract(pos[j])).length;
            m[i][j] = l;
            m[j][i] = l;
          }
        }
        return m;
      }

      calculateLessThanMatrix(ms, value, multi, type1, type2) {
    // console.log(ms);
        const m = [];
        for (let i = 0; i < this.N; i++) {
          const r = [];
          for (let j = 0; j < this.N; j++) {
            const v = ms[i][j];
            if (v < value + (value * this.size[i] * this.size[j] - value) * multi && this.types[i] === type1 && this.types[j] === type2) {
              r.push(v);
            } else {
              r.push(0.0);
            }
          }
          m.push(r);
        }
        return m;
      }

      tick() {
    // this.distMatrix = squareform(pdist(this.pos))
        this.distMatrix = this.calculateDistanceMatrix(this.pos);
    // console.log(this.distMatrix);
    // console.log("boid count "+this.N);

        const velAdd = this.applyRules();
    // const velAdd2 = this.applyPredatorRules();

        for (let i = 0; i < this.N; i++) {
          this.vel[i].addToThis(velAdd[i]);
      // this.vel[i].addToThis(velAdd2[i]);
        }

        for (let i = 0; i < this.N; i++) {
          if (this.types[i] === 0) {
            this.limitVec(this.vel[i], this.maxVel);
          }
        }

        for (let i = 0; i < this.N; i++) {
          if (this.types[i] === 1) {
            this.limitVec(this.vel[i], this.maxVel * 2.0);
          }
        }

        for (let i = 0; i < this.N; i++) { this.pos[i].addToThis(this.vel[i]); }

    // this.applyFoodRules();
        this.birthBoids();
        this.reaper();
      }

      applyRules() {
        const vel = [];
        for (let i = 0; i < this.N; i++) { vel.push(new Vec3(0.0, 0.0, 0.0)); }

    // # apply rule #1 - Separation
        if (this.rule1) {
      // D = this.distMatrix < this.minDist
          const D = this.calculateLessThanMatrix(this.distMatrix, this.minDist, 0.0, 0, 0);
          for (let i = 0; i < this.N; i++) {
            let sum = 0.0;
            const msum = new Vec3(0.0, 0.0, 0.0);
            for (let j = 0; j < this.N; j++) {
              sum += D[i][j];
              msum.addToThis(this.pos[j].multiple(D[i][j]));
            }
            const vel1 = this.pos[i].multiple(sum).subtract(msum);
            this.limitVec(vel1, this.maxRuleVel);
            vel[i].addToThis(vel1);
          }
        }
    // # different distance threshold
        const D2 = this.calculateLessThanMatrix(this.distMatrix, this.minDist * 4.0, 0.0, 0, 0);

    // # apply rule #2 - Alignment
        if (this.rule2) {
          for (let i = 0; i < this.N; i++) {
            const msum = new Vec3(0.0, 0.0, 0.0);
            for (let j = 0; j < this.N; j++) {
              msum.addToThis(this.vel[j].multiple(D2[i][j]));
            }
            this.limitVec(msum, this.maxRuleVel);
            vel[i].addToThis(msum);
          }
        }

    // # apply rule #3 - Cohesion

        if (this.rule3) {
          for (let i = 0; i < this.N; i++) {
            if (this.types[i] === 0) {
              const msum = new Vec3(0.0, 0.0, 0.0);
              for (let j = 0; j < this.N; j++) {
                msum.addToThis(this.pos[j].multiple(D2[i][j]));
              }
              const vel3 = msum.subtract(this.pos[i]);
              this.limitVec(vel3, this.maxRuleVel);
              vel[i].addToThis(vel3);
            }
          }
        }

    // # apply rul #4 stay close to center

        if (this.rule4) {
          const maxRadius = this.width / 4;
          for (let i = 0; i < this.N; i++) {
            if (this.types[i] == 0) {
              const n = this.center.subtract(this.pos[i]).length;
              if (n > maxRadius) {
                const nv = this.pos[i].subtract(this.center).multiple(1 / n);
                const nv2 = nv.multiple(2.0);
                const opposite = (nv.subtract(nv2)).multiple((n - maxRadius) / maxRadius * this.maxRuleVel);
                this.limitVec(opposite, this.maxRuleVel * 8.0);
                vel[i].addToThis(opposite);
              }
            }
          }
        }

    // # rule avoid trees

        if (this.rule5) {
      // vel4 = np.zeros((this.N,3))

          for (let i = 0; i < this.N; i++) {
        // if(this.types[i] === 0) {
            const vel4 = new Vec3(0.0, 0.0, 0.0);
          // for(let k=0; k<NB; k++){
        //    if(trees[j+k*w])
        //    {
            const direction = new Vec3(0.0, Math.min(4.25, this.pos[i].f[1]), 0.0).subtract(this.pos[i]);
            const n = direction.lengthSquared;
            const dist = (4.25 - this.pos[i].f[1]);
            if (n < Math.max(0.0, dist * dist)) {
              vel4.addToThis(direction.multiple(1.0 / n * 0.2 * (-n)));
            }
        //    }
        //  }
            this.limitVec(vel4, this.maxRuleVel * 2.0);
            vel[i].addToThis(vel4);
          }
        }


    // rule # 6 make random moves
        if (this.rule6) {
      // const angles = [];
      // const horiz = [];

          for (let i = 0; i < this.N; i++) {
            let angles = Math.PI * (Math.random() * 1.5 - 0.75);
            const x = this.vel[i].f[0];
            const y = this.vel[i].f[1];
            const z = this.vel[i].f[2];
            angles += Math.atan2(z, x);

            const horiz = Math.sqrt(x * x + z * z);

            const anglesup = Math.PI * (0.02 * Math.random() - 0.01) + Math.atan2(y, horiz);

            const vel6 = new Vec3(Math.cos(angles), Math.sin(anglesup), Math.sin(angles));
            this.limitVec(vel6, this.maxRuleVel / 3.0);
            vel[i].addToThis(vel6);
          }
        }


    // rule # 10 stay above certain height
    /*
        if (this.rule10) {
      // vel4 = np.zeros((this.N,3))

          for (let i = 0; i < this.N; i++) {
            if (this.types[i] === 0) {
              const vel4 = new Vec3(0.0, 0.0, 0.0);
          // for(let j=0; j< this.NB; j++) {
        //    const direction = this.balls[j].subtract(this.pos[i]);
      //      const n = direction.length;
              const n = grnd(this.pos[i].f[0], this.pos[i].f[1]) + 0.2 - this.pos[i].f[2];
            // if(this.types[i]==1)
            //  {
                // n=0;
              // n=ground(this.pos[i].f[0],this.pos[i].f[1])-this.pos[i].f[2];
              // n/=4.0;
            // console.log("g"+ground(this.pos[i].f[0],this.pos[i].f[1])+"n "+n+ "  "+ this.minHeight+ " "+this.pos[i].f[2]);
          // }
              if (n > 0) {
                vel4.addToThis(new Vec3(0.0, 0.0, n));
              }
          // }
              this.limitVec(vel4, this.maxRuleVel * 2.0);
              vel[i].addToThis(vel4);
            }
          }
        }
*/
    // rule # 11 stay below certain height
    /*
        if (this.rule11) {
      // vel4 = np.zeros((this.N,3))

          for (let i = 0; i < this.N; i++) {
            if (this.types[i] === 0) {
              const vel4 = new Vec3(0.0, 0.0, 0.0);
              const n = this.maxHeight - this.pos[i].f[1];
              // if(this.types[i]==1)
              //  n=0;//n=lake()-this.pos[i].f[2];
              if (n < 0) {
                vel4.addToThis(new Vec3(0.0, 0.0, n));
              }
              this.limitVec(vel4, this.maxRuleVel * 2.0);
              vel[i].addToThis(vel4);
            }
          }
        }
*/
    // rule # 12 kill all outside view area
    /*
        if (this.rule12) {
      // vel4 = np.zeros((this.N,3))

          for (let i = 0; i < this.N; i++) {
        // if(this.types[i] === 0) {
            const vel4 = new Vec3(0.0, 0.0, 0.0);
            const n = ((new Vec3(0, st * v, -0.5)).subtract(this.pos[i])).length();
            if (n > 20.0) {
              if (!this.markForDelete) { this.markForDelete = []; }
              this.markForDelete.push(i);
            }

        // }
          }
        }
*/

        return vel;
      }


}

    class BoidsRenderer {
      constructor(boids, gl) {
        this.gl = gl;
        this.boids = boids;
        this.vbuffer = null;
        this.vtbuffer = null;
        this.phbuffer = null;
        this.nbuffer = null;
        this.cbuffer = null;
        this.pbuffer = null;
        this.velbuffer = null;
        this.sBuffer = null;
        this.subdivision = 20;
        this.createShader(gl);
        this.createData();
        this.initData();
        this.createGLBuffers();
      }

      createShader(gl) {
        const vb = `
precision mediump float;
attribute vec3 aVel;
attribute vec3 aVert;
attribute float aPhase;
attribute float aVType;
attribute vec3 pos1;
attribute vec3 aNormal;
attribute vec3 aColor;
attribute float aSize;

uniform mat4 uMVMatrix;
uniform mat4 uPMatrix;
uniform mat4 uL;
uniform mat4 uC;

varying vec4 vCol;
varying vec4 aNormalInterp;
varying vec3 vertPos;
varying vec3 lightPos;


void main() {
float alpha = 1.0;
float PI = 3.14159265358979323846264;

      float theta = mod(atan(aVel.y,aVel.x)*180.0/PI, 360.0)*PI/180.0;
      float horiz = sqrt(aVel.x*aVel.x+aVel.y*aVel.y);
      float theta2 = mod(atan(aVel.z,horiz)*180.0/PI, 360.0)*PI/180.0;
      float ath = aVType*(((aPhase<0.5)?aPhase:1.0-aPhase)*1.5-0.45)*PI;
      float ath2 = aVType*(((aPhase<0.5)?aPhase:(1.0-aPhase))-0.25)*PI/4.0;

      mat4 anim = (abs(aVType)<1.5?mat4(
        vec4(1.0,0.0,0.0,0.0),
        vec4(0.0,cos(ath),sin(ath),0.0),

        vec4(0.0,-sin(ath),cos(ath),0.0),

        vec4(0.0,0.0,0.0,1.0)
      ):mat4(

        vec4(cos(ath2),sin(ath2),0.0,0.0),

        vec4(-sin(ath2),cos(ath2),0.0,0.0),
        vec4(0.0,0.0,1.0,0.0),
        vec4(0.0,0.0,0.0,1.0)
      ));

      mat4 scale =  mat4(
                 vec4(aSize,  0.0,0.0, 0.0),
                 vec4(0.0, aSize,0.0, 0.0),
                 vec4(0.0,  0.0, aSize, 0.0),
                 vec4(0.0,         0.0,         0.0, 1.0)
                );
mat4 rot =  mat4(
           vec4(cos(theta),  sin(theta),0.0, 0.0),
            vec4(-sin(theta), cos(theta),0.0, 0.0),
           vec4(0.0,                 0.0, 1.0, 0.0),
           vec4(0.0,         0.0,         0.0, 1.0)
          );
      mat4 rot2 = mat4(
                       vec4(cos(theta2),sin(theta2),0.0,0.0),
                       vec4(-sin(theta2),cos(theta2),0.0,0.0),
                       vec4(0.0,0.0,1.0,0.0),

                       vec4(0.0,0.0,0.0,1.0)
                      );
  vec4 pos2 =  rot*rot2*scale*anim*vec4(aVert, 1.0);
              vec3 newPos = pos2.xyz +pos1;
  gl_Position = uPMatrix*uC*uL* uMVMatrix * vec4(newPos, 1.0);
vCol = vec4(aColor.rgb, alpha);
  vertPos = (uMVMatrix * vec4(newPos, 1.0)).xyz;
  vec4 normal2 = rot*rot2*anim*vec4(aNormal,0.0);
  aNormalInterp = uMVMatrix * normal2;
  lightPos = (uMVMatrix* vec4(2.0,1.0,-2.0,0.0)).xyz;
}
`;
        const fb = `
precision mediump float;
varying vec4 vCol;
varying vec4  aNormalInterp;
varying vec3  vertPos;
varying vec3  lightPos;
const float screenGamma =2.2;

void main() {
   vec4 texCol = vec4(1.0,1.0,1.0,1.0);
   vec4 c = vCol;

    vec3 ambientColor = c.xyz/4.0;
  vec3 diffuseColor = c.xyz;
    vec3 normal = normalize(aNormalInterp.xyz);
  vec3 lightDir = normalize(lightPos);

    float lambertian = max(dot(lightDir,normal), 0.0);

    vec3 colorLinear = ambientColor +
                       lambertian * diffuseColor;;
    vec3 colorGammaCorrected = pow(colorLinear, vec3(1.0/screenGamma));
    gl_FragColor = vec4(colorGammaCorrected, 1.0);


}
`;

        this.program = ShaderLib.makeProgram(gl, vb, fb);
        const program = this.program;
        program.uPMatrix = gl.getUniformLocation(program, 'uPMatrix');
        program.uMVMatrix = gl.getUniformLocation(program, 'uMVMatrix');
        program.uC = gl.getUniformLocation(program, 'uC');
        program.uL = gl.getUniformLocation(program, 'uL');
        program.aVert = gl.getAttribLocation(program, 'aVert');
        program.aVType = gl.getAttribLocation(program, 'aVType');
        program.aPhase = gl.getAttribLocation(program, 'aPhase');
        program.aNormal = gl.getAttribLocation(program, 'aNormal');
        program.aVel = gl.getAttribLocation(program, 'aVel');
        program.pos1 = gl.getAttribLocation(program, 'pos1');
        program.aColor = gl.getAttribLocation(program, 'aColor');
        program.aSize = gl.getAttribLocation(program, 'aSize');
      }

      createData() {
        this.nvert = this.boids.maxN * 3 * 2;
        this.vertexData = new Float32Array(3 * this.nvert * 2);
        this.phaseData = new Float32Array(1 * this.nvert * 2);
        this.vertexTypeData = new Float32Array(1 * this.nvert * 2);
        this.normalData = new Float32Array(3 * this.nvert * 2);
        this.colorData = new Float32Array(3 * this.nvert * 2);
        this.positionData = new Float32Array(3 * this.nvert * 2);
        this.velocityData = new Float32Array(3 * this.nvert * 2);
        this.sizeData = new Float32Array(1 * this.nvert * 2);
      }

      createGLBuffers() {
        const glx = this.gl;

        this.vbuffer = glx.createBuffer();
        glx.bindBuffer(glx.ARRAY_BUFFER, this.vbuffer);
        glx.bufferData(glx.ARRAY_BUFFER, this.vertexData, glx.STATIC_DRAW);

        this.phbuffer = glx.createBuffer();
        glx.bindBuffer(glx.ARRAY_BUFFER, this.phbuffer);
        glx.bufferData(glx.ARRAY_BUFFER, this.phaseData, glx.DYNAMIC_DRAW);

        this.vtbuffer = glx.createBuffer();
        glx.bindBuffer(glx.ARRAY_BUFFER, this.vtbuffer);
        glx.bufferData(glx.ARRAY_BUFFER, this.vertexTypeData, glx.STATIC_DRAW);

        this.nbuffer = glx.createBuffer();
        glx.bindBuffer(glx.ARRAY_BUFFER, this.nbuffer);
        glx.bufferData(glx.ARRAY_BUFFER, this.normalData, glx.STATIC_DRAW);

        this.cbuffer = glx.createBuffer();
        glx.bindBuffer(glx.ARRAY_BUFFER, this.cbuffer);
        glx.bufferData(glx.ARRAY_BUFFER, this.colorData, glx.DYNAMIC_DRAW);

        this.pbuffer = glx.createBuffer();
        glx.bindBuffer(glx.ARRAY_BUFFER, this.pbuffer);
        glx.bufferData(glx.ARRAY_BUFFER, this.positionData, glx.DYNAMIC_DRAW);

        this.velbuffer = glx.createBuffer();
        glx.bindBuffer(glx.ARRAY_BUFFER, this.velbuffer);
        glx.bufferData(glx.ARRAY_BUFFER, this.velocityData, glx.DYNAMIC_DRAW);

        this.sbuffer = glx.createBuffer();
        glx.bindBuffer(glx.ARRAY_BUFFER, this.sbuffer);
        glx.bufferData(glx.ARRAY_BUFFER, this.sizeData, glx.DYNAMIC_DRAW);
      }

      draw(gl, canvas, pm, lm, cm, mm) {
    // console.log("draw");
        const program = this.program;
        gl.useProgram(program);
        gl.uniformMatrix4fv(program.uPMatrix, false, pm.m);
        gl.uniformMatrix4fv(program.uMVMatrix, false, mm.m);
        gl.uniformMatrix4fv(program.uL, false, lm.m);
        gl.uniformMatrix4fv(program.uC, false, cm.m);
        gl.bindBuffer(gl.ARRAY_BUFFER, this.vbuffer);
        gl.enableVertexAttribArray(program.aVert);
        gl.vertexAttribPointer(program.aVert, 3, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ARRAY_BUFFER, this.vtbuffer);
        gl.enableVertexAttribArray(program.aVType);
        gl.vertexAttribPointer(program.aVType, 1, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ARRAY_BUFFER, this.phbuffer);
        gl.enableVertexAttribArray(program.aPhase);
        gl.vertexAttribPointer(program.aPhase, 1, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ARRAY_BUFFER, this.nbuffer);
        gl.enableVertexAttribArray(program.aNormal);
        gl.vertexAttribPointer(program.aNormal, 3, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ARRAY_BUFFER, this.cbuffer);
        gl.enableVertexAttribArray(program.aColor);
        gl.vertexAttribPointer(program.aColor, 3, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ARRAY_BUFFER, this.pbuffer);
        gl.enableVertexAttribArray(program.pos1);
        gl.vertexAttribPointer(program.pos1, 3, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ARRAY_BUFFER, this.velbuffer);
        gl.enableVertexAttribArray(program.aVel);
        gl.vertexAttribPointer(program.aVel, 3, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ARRAY_BUFFER, this.sbuffer);
        gl.enableVertexAttribArray(program.aSize);
        gl.vertexAttribPointer(program.aSize, 1, gl.FLOAT, false, 0, 0);

        if (this.boids.N > 0) gl.drawArrays(gl.TRIANGLES, 0, this.boids.N * 3 * 2);
      }

      initData() {
        const s = 0.2;
        const w = 0.7;
        const h = 0.2;
        const quadV = [];
        const typeV = [];
        quadV.push(s);
        quadV.push(0);
        quadV.push(0);
        typeV.push(1);

        quadV.push(0);
        quadV.push(w);
        quadV.push(0);
        typeV.push(1);

        quadV.push(-s);
        quadV.push(0);
        quadV.push(0);
        typeV.push(1);

        quadV.push(s);
        quadV.push(0);
        quadV.push(0);
        typeV.push(-1);

        quadV.push(-s);
        quadV.push(0);
        quadV.push(0);
        typeV.push(-1);

        quadV.push(0);
        quadV.push(-w);
        quadV.push(0);
        typeV.push(-1);


        const normalV = [];
        normalV.push(0);
        normalV.push(1);
        normalV.push(0);

        normalV.push(0);
        normalV.push(1);
        normalV.push(0);

        normalV.push(0);
        normalV.push(1);
        normalV.push(0);

        normalV.push(0);
        normalV.push(1);
        normalV.push(0);

        normalV.push(0);
        normalV.push(1);
        normalV.push(0);

        normalV.push(0);
        normalV.push(1);
        normalV.push(0);

        let offset = 0;
        const nv = 3 * 2;
        for (let i = 0; i < this.boids.maxN * 2; i++) {
          for (let j = 0; j < nv; j++) {
            this.vertexData[offset++] = quadV[j * 3 + 0];
            this.vertexData[offset++] = quadV[j * 3 + 1];
            this.vertexData[offset++] = quadV[j * 3 + 2];
          }
        }
        offset = 0;
        for (let i = 0; i < this.boids.maxN * 2; i++) {
          for (let j = 0; j < nv; j++) {
            this.normalData[offset++] = normalV[j * 3 + 0];
            this.normalData[offset++] = normalV[j * 3 + 1];
            this.normalData[offset++] = normalV[j * 3 + 2];
          }
        }

        offset = 0;
        for (let i = 0; i < this.boids.maxN * 2; i++) {
          const boidphase = Math.random();
          for (let j = 0; j < nv; j++) {
            this.vertexTypeData[offset] = typeV[j];
            this.phaseData[offset++] = boidphase;
          }
        }

        this.updateVertices();
    // console.log("buffers "+ this.positionData);
      }

      updateVertices() {
    // console.log("updatevertive");
        let offset = 0;
        const nv = 3 * 2;
        for (let i = 0; i < this.boids.N; i++) {
          for (let j = 0; j < nv; j++) {
            this.positionData[offset++] = this.boids.pos[i].f[0];
            this.positionData[offset++] = this.boids.pos[i].f[1];
            this.positionData[offset++] = this.boids.pos[i].f[2];
          }
        }
        offset = 0;
        for (let i = 0; i < this.boids.N; i++) {
          for (let j = 0; j < nv; j++) {
        // const v =this.boids.vel[i].f;
            let v2 = 0.04;// //Math.max(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]*1000,0.04);
            if (this.boids.type == 1) v2 = 0.05;// Math.max(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]*10000,0.05);
            this.phaseData[offset] = (this.phaseData[offset++] + v2) % 1.0;
          }
        }
        offset = 0;
        for (let i = 0; i < this.boids.N; i++) {
          for (let j = 0; j < nv; j++) {
            this.velocityData[offset++] = this.boids.vel[i].f[0];
            this.velocityData[offset++] = this.boids.vel[i].f[1];
            this.velocityData[offset++] = this.boids.vel[i].f[2];
          }
        }
        offset = 0;
        for (let i = 0; i < this.boids.N; i++) {
          for (let j = 0; j < nv; j++) {
            this.colorData[offset++] = this.boids.color[i].f[0];
            this.colorData[offset++] = this.boids.color[i].f[1];
            this.colorData[offset++] = this.boids.color[i].f[2];
          }
        }

        offset = 0;
        for (let i = 0; i < this.boids.N; i++) {
          for (let j = 0; j < nv; j++) {
            this.sizeData[offset++] = this.boids.size[i];
          }
        }

        const gl = this.gl;
        if (this.pbuffer !== null) {
          gl.bindBuffer(gl.ARRAY_BUFFER, this.phbuffer);
          gl.bufferData(gl.ARRAY_BUFFER, this.phaseData, gl.STATIC_DRAW);

          gl.bindBuffer(gl.ARRAY_BUFFER, this.pbuffer);
          gl.bufferData(gl.ARRAY_BUFFER, this.positionData, gl.STATIC_DRAW);

          gl.bindBuffer(gl.ARRAY_BUFFER, this.velbuffer);
          gl.bufferData(gl.ARRAY_BUFFER, this.velocityData, gl.STATIC_DRAW);

          gl.bindBuffer(gl.ARRAY_BUFFER, this.sbuffer);
          gl.bufferData(gl.ARRAY_BUFFER, this.sizeData, gl.STATIC_DRAW);

          gl.bindBuffer(gl.ARRAY_BUFFER, this.cbuffer);
          gl.bufferData(gl.ARRAY_BUFFER, this.colorData, gl.STATIC_DRAW);
        }
      }
}

    class TestBoids {
      constructor(gl) {
        this.boids = new Boids(300, 0);
        for (let i = 0; i < 50; i++) {
          this.boids.addBoids((Math.random() - 0.5) * 10, (Math.random() - 0.5) * 10, Math.random() * 8);
        }
        for (let i = 0; i < 1000; i++) {
          this.boids.tick();
        }
        // let boids2 = new Boids(300,1);
        this.boidsRenderer = new BoidsRenderer(this.boids, gl);
      }

      tick(t) {
        this.boids.tick(t);
        this.boidsRenderer.updateVertices();
      }

      draw(gl, canvas, pm, vm, cm, mm) {
        this.boidsRenderer.draw(gl, canvas, pm, vm, cm, mm);
      }
    }

    BoidsLib.Boids = Boids;
    BoidsLib.BoidsRenderer = BoidsRenderer;
    BoidsLib.TestBoids = TestBoids;

    return BoidsLib;
  }
// define globally if it doesn't already exist
  if (typeof (BoidsLib) === 'undefined') {
    const mwindow = window;
    mwindow.BoidsLib = defineLibrary();
  }
}(window)); // eslint-disable-line no-undef
