(function capsule(window) {
  function defineLibrary() {
    const SwarmLib = {};
    const Vec3 = VemaLib.Vec3;
    const Matrix4 = VemaLib.Matrix4;
    function bc(p0, p1, p2, p3, t) {
      return (1 - t) * (1 - t) * (1 - t) * p0 + 3 * (1 - t) * (1 - t) * t * p1 + 3 * (1 - t) * t * t * p2 + t * t * t * p3;
    }

    function bcd(p0, p1, p2, p3, t) {
      return 3 * (1 - t) * (1 - t) * (p1 - p0) + 6 * (1 - t) * t * (p2 - p1) + 3 * t * t * (p3 - p2);
    }

    class ButterflyRenderer {
      constructor(boids, gl, canvas, dataset) {
        this.dataset = dataset;
        this.gl = gl;
        this.boids = boids;

        this.wingSpread = 0;
        this.spreadSpeed = 1.6;
        this.scrollSpeed = 0;
        this.roll = 0;
        this.colorAdd = [0, 0, 0];
        this.stage = 0;
        this.elem = null;
        this.cb = null;
        this.noBF = 6;
        this.N = Math.floor(this.boids.N / this.noBF);
        this.bflies = [];

        this.colors = [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0];
      }

      init(gl, canvas, textElem, callback) {
        if (this.cb === null) {
          this.cb = callback;
        }
        if (this.stage === 0) {
          this.stage += 1;
          textElem.innerHTML += "<p id='bfly'>ButterFLies Creation 0%</p>";
          this.elem = document.getElementById('bfly');
          this.createShader(gl);
          setTimeout(this.init(gl, canvas, textElem), 10);
          return;
        }
        if (this.stage === 1) {
          this.stage += 1;
          this.elem.innerHTML = 'Butterflies Creation 10%';
          for (let i = 0; i < this.noBF; i++) {
            this.initDataBuffers(gl, i);
            this.makeOneFly(gl, i);
          }
          /*
          this.makeOneFly(gl, 1);
          this.makeOneFly(gl, 2);
          this.makeOneFly(gl, 3);
          */
          setTimeout(this.init(gl, canvas, textElem), 10);
          return;
        }
        if (this.stage < 10) {
          this.stage += 1;
          this.elem.innerHTML = `Butterflies Creation ${(100 * (this.stage - 2) / 8).toFixed(2)}%`;
          for (let i = 0; i < this.noBF; i++) {
            this.makeDataBuffers(gl, this.stage - 3, i);
          }
          setTimeout(this.init(gl, canvas, textElem), 500);
          return;
        }
        if (this.stage === 10) {
          this.stage += 1;
          for (let i = 0; i < this.noBF; i++) {
            this.createGLBuffers(gl, i);
          }
          setTimeout(this.init(gl, canvas, textElem), 500);
          return;
        }
        this.cb();
      }

      tick(t) {
        this.roll += this.scrollSpeed;
        if (this.roll > 255.0) this.roll = 0.0;
        if (this.roll < 0.0) this.roll = 255.0;
      }

      createShader(gl) {
        const vb = `#version 300 es
    precision highp float;
    in vec4 vPosition;
    in vec2 multiTexCoord0;
    in vec2 aWingOffset;
    in float aIndex;
    in vec3 aNormal;
    uniform sampler2D uTextureData;
    uniform vec2 uTextureDataSize;
    out vec2 texCoord0;
    out vec3 normal;
    out float offset;
    uniform mat4 projection;
    uniform mat4 camera;
    uniform mat4 view;
    uniform mat4 model;
    uniform float wingSpeed;
    const float cSize = 0.05;
    const float PI = 3.14159265358979323846264;
    void main() {
      vec4 data0 = texture(uTextureData,vec2((aIndex+0.5)*1.0/uTextureDataSize.x,0.5*1.0/uTextureDataSize.y));
      vec4 data1 = texture(uTextureData,vec2((aIndex+0.5)*1.0/uTextureDataSize.x,1.5*1.0/uTextureDataSize.y));

      float theta = -atan(data1.z,data1.x);
      float horiz = sqrt(data1.x*data1.x+data1.z*data1.z);
      float theta3 = -atan(data1.y,horiz);

      float wings = sin((data1.w-aWingOffset.x*wingSpeed))*aWingOffset.y;
      offset = aWingOffset.y;
      float ws = ((vPosition.z>0.0)?((wings+0.5)):(-(wings+0.5)));
      mat4 trans = mat4(1.0,0.0,0.0,0.0,
                        0.0,1.0,0.0,0.0,
                        0.0,0.0,1.0,0.0,
                        0.0,-sin(data1.w)*0.3*cSize,0.0,1.0);
      mat4 wsm;
      wsm[0] = vec4(1.0, 0.0, 0.0, 0.0);
      wsm[1] = vec4(0.0,cos(ws), -sin(ws), 0.0);

      wsm[2] = vec4(0.0, sin(ws), cos(ws), 0.0);
      wsm[3] = vec4(0.0, 0.0, 0.0, 1.0);
      texCoord0 = multiTexCoord0;
      mat4 nrm;

      nrm[0] = vec4(sin(-ws), cos(-ws),0.0,0.0);
      nrm[1] = vec4(-cos(-ws),sin(-ws),0.0,0.0);
      nrm[2] = vec4(0.0,0.0, 1.0,0.0);
      nrm[3] = vec4(0.0,0.0,0.0,1.0);


      mat4 scale =  mat4(
                 vec4(cSize,  0.0,0.0, 0.0),
                 vec4(0.0, cSize,0.0, 0.0),
                 vec4(0.0,  0.0, cSize, 0.0),
                 vec4(0.0,         0.0,         0.0, 1.0)
                );
            mat4 rotZ =  mat4(
                   vec4(cos(theta3),-sin(theta3),0.0,0.0),
                    vec4(sin(theta3),cos(theta3),0.0, 0.0),
                    vec4(0.0,                 0.0,1.0, 0.0),
                   vec4(0.0,         0.0,         0.0, 1.0)
                  );

      mat4 rot =  mat4(
             vec4(cos(theta),0.0,   -sin(theta),0.0),
              vec4(0.0,                 1.0,0.0, 0.0),
              vec4(sin(theta),0.0, cos(theta), 0.0),
             vec4(0.0,         0.0,         0.0, 1.0)
            );
      vec4 pos2 =  rot*rotZ*scale*wsm*vPosition;
      vec3 newPos = pos2.xyz +data0.xyz;

      vec3 norman = normalize((nrm*wsm*vec4(aNormal,1.0)).xyz);
      normal = (view*model*rot*rotZ*wsm*vec4(norman,1.0)).xyz;
      gl_Position = (projection * camera * view *trans* model * vec4(newPos,1.0));
    }
    `;
        const fb = `#version 300 es
    precision highp float;
    uniform mat4 uRotMatrix;
    in vec2 texCoord0;
    in vec3 normal;
    in float offset;
    uniform vec2 centerOffset;
    uniform float zoom;
    uniform float colorroll;
    uniform int colorCount;
    uniform vec3 colors[9];
    uniform vec3 colorAdd;
    out vec4 fragColor;
    const int iterations =255;
    highp float mandel(vec2 tc) {
       highp vec2 c = (zoom*(uRotMatrix*vec4(tc ,0.0,0.0)).xy+centerOffset);
       highp vec2 z = c;
       for (int n = 0; n < iterations; n++) {
           z = vec2(z.x * z.x - z.y * z.y, 2.0 * z.x * z.y) + c;
           if (length(z) > 4.0) return (float(n) + 1.0 - log(log(length(z))) / log(2.0));
       }
       return 0.0;
    }

    void main() {
        highp float n= mandel(texCoord0);
        vec3 color = vec3(0.0,0.0,0.0);
        if (n == 0.0|| offset <0.5){
          color = vec3(0.1, 0.1, 0.1);
        }else{
           n+=colorroll*50.0;
           float value = fract(n*0.02);
           int index0 = int(floor(value*float(colorCount)));
           float f = value*float(colorCount)-float(index0);
           for(int i=0; i<9; i++){
              vec3 c0 = colors[i];
              vec3 c1 = (i<colorCount-1)?colors[i+1]:colors[0];
              color = (i==index0)?mix(c0,c1,f):color;
           }
         }
         vec3 c = color+colorAdd*0.2;

         vec3 ambientColor = c.xyz/4.0;
         vec3 diffuseColor = c.xyz;
         vec3 norm = normalize(normal);
         vec3 lightDir = normalize(vec3(2.0,20.0,3.0));

         float lambertian = max(abs(dot(norm,lightDir)), 0.0);

         vec4 colorLinear = vec4(ambientColor +
                            lambertian * diffuseColor,1.);
         fragColor = colorLinear;

    }
    `;

        this.program = ShaderLib.makeProgram(gl, vb, fb);
        const program = this.program;
        this.params = [];
        this.params.aIndex = gl.getAttribLocation(program, 'aIndex');
        this.params.aPosition = gl.getAttribLocation(program, 'vPosition');
        this.params.aTexCoordinate = gl.getAttribLocation(program, 'multiTexCoord0');
        this.params.aWingOffset = gl.getAttribLocation(program, 'aWingOffset');
        this.params.aNormal = gl.getAttribLocation(program, 'aNormal');
        this.params.uRotMatrix = gl.getUniformLocation(program, 'uRotMatrix');
        this.params.colorroll = gl.getUniformLocation(program, 'colorroll');
        this.params.zoom = gl.getUniformLocation(program, 'zoom');
        this.params.xy = gl.getUniformLocation(program, 'centerOffset');
        this.params.projection = gl.getUniformLocation(program, 'projection');
        this.params.camera = gl.getUniformLocation(program, 'camera');
        this.params.model = gl.getUniformLocation(program, 'model');
        this.params.view = gl.getUniformLocation(program, 'view');
        this.params.wingSpeed = gl.getUniformLocation(program, 'wingSpeed');
        this.params.colorAdd = gl.getUniformLocation(program, 'colorAdd');
        this.params.colorCount = gl.getUniformLocation(program, 'colorCount');
        this.params.colors = gl.getUniformLocation(program, 'colors');
        this.params.uTextureData = gl.getUniformLocation(program, 'uTextureData');
        this.params.uTextureDataSize = gl.getUniformLocation(program, 'uTextureDataSize');
      }

      draw(gl, canvas, pm, lm, cm, mm) {
        for (let i = 0; i < this.noBF; i++) {
          this.drawOnTypeOfBflies(gl, canvas, pm, lm, cm, mm, i);
        }
      }
      drawOnTypeOfBflies(gl, canva, pm, lm, cm, mm, indexOfData) {
        const program = this.program;
        gl.useProgram(program);

        const bfly = this.bflies[indexOfData];

        gl.bindBuffer(gl.ARRAY_BUFFER, bfly.vbuffer);
        gl.enableVertexAttribArray(this.params.aPosition);
        gl.vertexAttribPointer(this.params.aPosition, 3, gl.FLOAT, false, 0, 0);
        gl.bindBuffer(gl.ARRAY_BUFFER, bfly.nbuffer);
        gl.enableVertexAttribArray(this.params.aNormal);
        gl.vertexAttribPointer(this.params.aNormal, 3, gl.FLOAT, false, 0, 0);
        gl.bindBuffer(gl.ARRAY_BUFFER, bfly.tbuffer);
        gl.enableVertexAttribArray(this.params.aTexCoordinate);
        gl.vertexAttribPointer(this.params.aTexCoordinate, 2, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ARRAY_BUFFER, bfly.obuffer);
        gl.enableVertexAttribArray(this.params.aWingOffset);
        gl.vertexAttribPointer(this.params.aWingOffset, 2, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ARRAY_BUFFER, bfly.inbuffer);
        gl.enableVertexAttribArray(this.params.aIndex);
        gl.vertexAttribPointer(this.params.aIndex, 1, gl.FLOAT, false, 0, 0);

        const r = new Matrix4();
        r.rotationZMatrix(bfly.rotate);
        gl.uniformMatrix4fv(this.params.uRotMatrix, false, new Float32Array(r.m));
        gl.uniform2f(this.params.xy, bfly.offsetx, bfly.offsety);
        gl.uniform2f(this.params.uTextureDataSize, this.boids.N, 8);
        gl.uniform1f(this.params.zoom, bfly.zoom);
        gl.uniform1f(this.params.colorroll, this.roll);
        gl.uniform1i(this.params.colorCount, Math.floor(bfly.colors.length / 3));
        gl.uniform3fv(this.params.colors, new Float32Array(bfly.colors));
        gl.uniform3f(this.params.colorAdd, this.colorAdd[0], this.colorAdd[1], this.colorAdd[2]);
        gl.uniformMatrix4fv(this.params.projection, false,
                new Float32Array(pm.m));
        gl.uniformMatrix4fv(this.params.camera, false, new Float32Array(cm.m));
        gl.uniformMatrix4fv(this.params.model, false, new Float32Array(mm.m));
        gl.uniformMatrix4fv(this.params.view, false, new Float32Array(lm.m));
        gl.uniform1f(this.params.wingSpeed, this.spreadSpeed);

        gl.activeTexture(this.boids.ofb.textureUnit.unit);
        gl.bindTexture(gl.TEXTURE_2D, this.boids.ofb.texture);
        gl.uniform1i(this.params.uTextureData, this.boids.ofb.textureUnit.number);

        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, bfly.ibuffer);
        gl.drawElements(gl.TRIANGLE_STRIP, bfly.pc, gl.UNSIGNED_INT, 0);
      }

      initDataBuffers(gl, indexOfData) {
        const bfly = this.bflies[indexOfData] = {};
        bfly.offsetbase = 0;
        if (bfly.vbuffer) {
          gl.deleteBuffer(bfly.vbuffer);
          bfly.vbuffer = null;
        }
        if (bfly.tbuffer) {
          gl.deleteBuffer(bfly.tbuffer);
          bfly.tbuffer = null;
        }
        if (bfly.obuffer) {
          gl.deleteBuffer(bfly.obuffer);
          bfly.obuffer = null;
        }
        if (bfly.inbuffer) {
          gl.deleteBuffer(bfly.inbuffer);
          bfly.inbuffer = null;
        }
        if (bfly.nbuffer) {
          gl.deleteBuffer(bfly.nbuffer);
          bfly.nbuffer = null;
        }
        if (bfly.ibuffer) {
          gl.deleteBuffer(bfly.ibuffer);
          bfly.ibuffer = null;
        }

        bfly.vbt = [];
        bfly.tbt = [];
        bfly.obt = [];
        bfly.nbt = [];
      }

      makeOneFly(gl, indexOfData) {
        const bfly = this.bflies[indexOfData];
        const dataset = this.dataset[indexOfData];
        bfly.rotate = dataset.rotate;
        bfly.offsetx = dataset.offsetx;
        bfly.offsety = dataset.offsety;
        bfly.zoom = dataset.zoom;
        bfly.colors = dataset.colors;
        bfly.vbl = [];
        bfly.vbr = [];
        bfly.tbl = [];
        bfly.tbr = [];
        bfly.obl = [];
        bfly.obr = [];
        bfly.nbl = [];
        bfly.nbr = [];


        bfly.c = dataset.wing.upper.points.length / 2;

        bfly.dc = 3;

        for (let i = 0; i < bfly.c; i++) {
          const px = dataset.wing.upper.points[i * 2];
          const py = -dataset.wing.upper.points[i * 2 + 1];
          const p = { x: px, y: py };
          const tp = { x: px - 0.5, y: (py + 0.5) };
          const woffset = i / bfly.c;
          if (i < bfly.c - 1) {
            const px2 = dataset.wing.upper.points[(i + 1) * 2];
            const py2 = -dataset.wing.upper.points[(i + 1) * 2 + 1];
            const p2 = { x: px2, y: py2 };
            const tp2 = { x: px2 - 0.5, y: (py2 + 0.5) };
            const cpp0 = { x: dataset.wing.upper.controlPoints[i * 4], y: -dataset.wing.upper.controlPoints[i * 4 + 1] };
            const cpp1 = { x: dataset.wing.upper.controlPoints[i * 4 + 2], y: -dataset.wing.upper.controlPoints[i * 4 + 3] };
            const cp0 = { x: cpp0.x, y: cpp0.y };
            const cp1 = { x: cpp1.x, y: cpp1.y };
            const tcp0 = { x: cpp0.x - 0.5, y: (cpp0.y + 0.5) };
            const tcp1 = { x: cpp1.x - 0.5, y: (cpp1.y + 0.5) };

            for (let j = 0; j < bfly.dc; j++) {
              const woffset2 = woffset + 1 / bfly.c * j / (bfly.dc - 1);
              const st = j / (bfly.dc - 1);

              bfly.vbl.push(bc(p.y, cp0.y, cp1.y, p2.y, st));
              bfly.vbl.push(0.0);
              bfly.vbl.push(bc(p.x, cp0.x, cp1.x, p2.x, st));

              bfly.vbr.push(bc(p.y, cp0.y, cp1.y, p2.y, st));
              bfly.vbr.push(0.0);
              bfly.vbr.push(-bc(p.x, cp0.x, cp1.x, p2.x, st));

              bfly.tbl.push(bc(tp.x, tcp0.x, tcp1.x, tp2.x, st));
              bfly.tbl.push(bc(tp.y, tcp0.y, tcp1.y, tp2.y, st));
              bfly.tbr.push(bc(tp.x, tcp0.x, tcp1.x, tp2.x, st));
              bfly.tbr.push(bc(tp.y, tcp0.y, tcp1.y, tp2.y, st));
              bfly.obl.push(woffset2);
              bfly.obl.push(1.0);
              bfly.obr.push(woffset2);
              bfly.obr.push(1.0);

              bfly.nbl.push(0.0);
              bfly.nbl.push(1.0);
              bfly.nbl.push(0.0);

              bfly.nbr.push(0.0);
              bfly.nbr.push(1.0);
              bfly.nbr.push(0.0);
            }
          }
        }


        bfly.lvbl = [];
        bfly.lvbr = [];
        bfly.ltbl = [];
        bfly.ltbr = [];
        bfly.lobl = [];
        bfly.lobr = [];
        bfly.lnbl = [];
        bfly.lnbr = [];

        bfly.lc = dataset.wing.lower.points.length / 2;

                // const dc = 20;

        for (let i = 0; i < bfly.lc; i++) {
          const px = dataset.wing.lower.points[i * 2];
          const py = -dataset.wing.lower.points[i * 2 + 1];
          const p = { x: px, y: py };
          const tp = { x: px - 0.5, y: (py + 0.5) };
          const woffset = i / bfly.lc;
          if (i < bfly.lc - 1) {
            const px2 = dataset.wing.lower.points[(i + 1) * 2];
            const py2 = -dataset.wing.lower.points[(i + 1) * 2 + 1];
            const p2 = { x: px2, y: py2 };
            const tp2 = { x: px2 - 0.5, y: (py2 + 0.5) };
            const cpp0 = { x: dataset.wing.lower.controlPoints[i * 4], y: -dataset.wing.lower.controlPoints[i * 4 + 1] };
            const cpp1 = { x: dataset.wing.lower.controlPoints[i * 4 + 2], y: -dataset.wing.lower.controlPoints[i * 4 + 3] };
            const cp0 = { x: cpp0.x, y: cpp0.y };
            const cp1 = { x: cpp1.x, y: cpp1.y };
            const tcp0 = { x: cpp0.x - 0.5, y: (cpp0.y + 0.5) };
            const tcp1 = { x: cpp1.x - 0.5, y: (cpp1.y + 0.5) };

            for (let j = 0; j < bfly.dc; j++) {
              const woffset2 = 0.5 + woffset + 1 / bfly.lc * j / (bfly.dc - 1);
              const st = j / (bfly.dc - 1);

              bfly.lvbl.push(bc(p.y, cp0.y, cp1.y, p2.y, st));
              bfly.lvbl.push(0.0);
              bfly.lvbl.push(bc(p.x, cp0.x, cp1.x, p2.x, st));

              bfly.lvbr.push(bc(p.y, cp0.y, cp1.y, p2.y, st));
              bfly.lvbr.push(0.0);
              bfly.lvbr.push(-bc(p.x, cp0.x, cp1.x, p2.x, st));

              bfly.ltbl.push(bc(tp.x, tcp0.x, tcp1.x, tp2.x, st));
              bfly.ltbl.push(bc(tp.y, tcp0.y, tcp1.y, tp2.y, st));
              bfly.ltbr.push(bc(tp.x, tcp0.x, tcp1.x, tp2.x, st));
              bfly.ltbr.push(bc(tp.y, tcp0.y, tcp1.y, tp2.y, st));
              bfly.lnbl.push(0.0);
              bfly.lnbl.push(1.0);
              bfly.lnbl.push(0.0);

              bfly.lnbr.push(0.0);
              bfly.lnbr.push(1.0);
              bfly.lnbr.push(0.0);

              bfly.lobl.push(woffset2);
              bfly.lobl.push(1.0);
              bfly.lobr.push(woffset2);
              bfly.lobr.push(1.0);
            }
          }
        }
        this.makeBody(gl, indexOfData);
        this.makeAntennae(gl, indexOfData);
        bfly.offset = (bfly.dc * (bfly.c - 1));
        bfly.loffset = (bfly.dc * (bfly.lc - 1));
        bfly.boffset = (bfly.dcbody * bfly.drbody * (bfly.cbody - 1));
        bfly.totaloffset = bfly.offset * 2 + bfly.loffset * 2 + bfly.boffset + bfly.antennaeoffset * 2;
        bfly.indexcount = bfly.totaloffset * 2 + 14;

        bfly.vbt = bfly.vbl.concat(bfly.vbr).concat(bfly.lvbl).concat(bfly.lvbr).concat(bfly.vbbody).concat(bfly.vba);
        bfly.tbt = bfly.tbl.concat(bfly.tbr).concat(bfly.ltbl).concat(bfly.ltbr).concat(bfly.tbbody).concat(bfly.tba);
        bfly.obt = bfly.obl.concat(bfly.obr).concat(bfly.lobl).concat(bfly.lobr).concat(bfly.obbody).concat(bfly.oba);
        bfly.nbt = bfly.nbl.concat(bfly.nbr).concat(bfly.lnbl).concat(bfly.lnbr).concat(bfly.nbbody).concat(bfly.nba);
        bfly.ib = new Uint32Array(bfly.indexcount * this.N);
      }

      makeAntennae(gl, indexOfData) {
        const vbla = [];
        const tbla = [];
        const vbra = [];
        const tbra = [];
        const obra = [];
        const obla = [];
        const nbra = [];
        const nbla = [];
        const bfly = this.bflies[indexOfData];

        bfly.vba = [];
        bfly.tba = [];
        bfly.oba = [];
        bfly.nba = [];

        const dr = 8;
        const dc = 4;
        const dataset = this.dataset[indexOfData];

        const rootpoint = { x: dataset.body.points[0], y: dataset.body.points[1] };

        // const points = [{ x: 0.00, y: 0.0 }, { x: 0.2, y: -0.3 }];
        const c = dataset.antennae.points.length / 3;
        // const anglex = 0.2;
        // const angley = 0.1;
        const awidth = 0.005;

        // const rotX = new Matrix4();
        // const rotY = new Matrix4();
        const trans = new Matrix4();


        // rotX.rotationXMatrix(anglex);
        // rotY.rotationYMatrix(angley);
        trans.translate(rootpoint.x, -rootpoint.y, 0);
        // const rot = rotY.multiple(rotX.multiple(trans));
        for (let i = 0; i < c - 1; i++) {
          const pp = { x: dataset.antennae.points[i * 3], y: dataset.antennae.points[i * 3 + 1], z: dataset.antennae.points[i * 3 + 2] };
          const pp2 = { x: dataset.antennae.points[(i + 1) * 3], y: dataset.antennae.points[(i + 1) * 3 + 1], z: dataset.antennae.points[(i + 1) * 3 + 2] };
          const p = { x: pp.x, y: pp.y, z: pp.z };
          const p2 = { x: pp2.x, y: pp2.y, z: pp2.z };
          const tp = { x: pp.x - 0.5, y: (pp.y - 0.5) };
          const cpp0 = { x: dataset.antennae.controlPoints[i * 6], y: dataset.antennae.controlPoints[i * 6 + 1], z: dataset.antennae.controlPoints[i * 6 + 2] };
          const cpp1 = { x: dataset.antennae.controlPoints[i * 6 + 3], y: dataset.antennae.controlPoints[i * 6 + 4], z: dataset.antennae.controlPoints[i * 6 + 5] };
          const cp0 = { x: cpp0.x, y: cpp0.y, z: cpp0.z };
          const cp1 = { x: cpp1.x, y: cpp1.y, z: cpp1.z };
          for (let j = 0; j < dc; j++) {
            const ta = j / (dc - 1);
            const ap = new Vec3(bc(p.x, cp0.x, cp1.x, p2.x, ta), bc(p.y, cp0.y, cp1.y, p2.y, ta), bc(p.z, cp0.z, cp1.z, p2.z, ta));
            const tb = (j + 1) / (dc - 1);
            const bp = new Vec3(bc(p.x, cp0.x, cp1.x, p2.x, tb), bc(p.y, cp0.y, cp1.y, p2.y, tb), bc(p.z, cp0.z, cp1.z, p2.z, tb));
            const norm = new Vec3(0.0, 0.0, 1.0);
            const dir = bp.subtract(ap);
            const xcomp = dir.cross(norm).normalized();
            const zcomp = xcomp.cross(dir).normalized();
            for (let k = 0; k < dr; k++) {
              const a = (k / (dr - 1)) * Math.PI * 2;
              // const vp = new Vec3(Math.sin(a) * awidth, p.y, Math.cos(a) * awidth);
              const vp = ap.add(xcomp.multiple(Math.sin(a) * awidth)).add(zcomp.multiple(Math.cos(a) * awidth));
              const v = trans.multipleV(vp);
              const n = xcomp.normalized();
              vbla.push(v.f[1]);
              vbla.push(v.f[2]);
              vbla.push(v.f[0]);
              tbla.push(0.75);
              tbla.push(0.125);
              vbra.push(v.f[1]);
              vbra.push(v.f[2]);
              vbra.push(-v.f[0]);
              tbra.push(0.75);
              tbra.push(0.125);
              nbla.push(n.f[1]);
              nbla.push(n.f[2]);
              nbla.push(n.f[0]);
              nbra.push(n.f[1]);
              nbra.push(n.f[2]);
              nbra.push(-n.f[0]);
              obla.push(0.75);
              obla.push(0.125);
              obra.push(0.75);
              obra.push(0.125);
            }
          }
        }

        bfly.antennaeoffset = (c - 1) * (dc) * dr;
        bfly.dca = dc;
        bfly.dra = dr;
        bfly.ca = c;


        bfly.vba = vbla.concat(vbra);
        bfly.tba = tbla.concat(tbra);
        bfly.oba = obla.concat(obra);
        bfly.nba = nbla.concat(nbra);

        // this.iba = ibla.concat(ibra);
      }

      makeBody(gl, indexOfData) {
        const bfly = this.bflies[indexOfData];

        bfly.vbbody = [];
        bfly.tbbody = [];
        bfly.obbody = [];
        bfly.nbbody = [];
        const dataset = this.dataset[indexOfData];
        bfly.cbody = dataset.body.points.length / 2;
        bfly.drbody = 8;
        bfly.dcbody = 4;

        for (let i = 0; i < bfly.cbody - 1; i++) {
          const px = dataset.body.points[i * 2];
          const py = dataset.body.points[i * 2 + 1];
          const p = { x: px, y: -py };
          const tp = { x: py - 0.5, y: -(px - 0.5) };
          const px2 = dataset.body.points[(i + 1) * 2];
          const py2 = dataset.body.points[(i + 1) * 2 + 1];
          const p2 = { x: px2, y: -py2 };
          const tp2 = { x: px2 - 0.5, y: -(py2 - 0.5) };
          const cpp0 = { x: dataset.body.controlPoints[i * 4], y: dataset.body.controlPoints[i * 4 + 1] };
          const cpp1 = { x: dataset.body.controlPoints[i * 4 + 2], y: dataset.body.controlPoints[i * 4 + 3] };
          const cp0 = { x: cpp0.x, y: -cpp0.y };
          const cp1 = { x: cpp1.x, y: -cpp1.y };
          const tcp0 = { x: cpp0.x - 0.5, y: -(cpp0.y - 0.5) };
          const tcp1 = { x: cpp1.x - 0.5, y: -(cpp1.y - 0.5) };

          for (let j = 0; j < bfly.dcbody; j++) {
            const st = j / (bfly.dcbody - 1);
            const pq = { x: bc(p.x, cp0.x, cp1.x, p2.x, st),
              y: bc(p.y, cp0.y, cp1.y, p2.y, st) };
            const pqd = { x: bcd(p.x, cp0.x, cp1.x, p2.x, st),
              y: bcd(p.y, cp0.y, cp1.y, p2.y, st) };
            const tpq = { x: bc(tp.x, tcp0.x, tcp1.x, tp2.x, st),
              y: bc(tp.y, tcp0.y, tcp1.y, tp2.y, st) };
            for (let k = 0; k < bfly.drbody; k++) {
              const a = (k / (bfly.drbody - 1)) * Math.PI * 2;
              bfly.vbbody.push(pq.y);
              bfly.vbbody.push(Math.cos(a) * pq.x);
              bfly.vbbody.push(Math.sin(a) * pq.x);

              bfly.tbbody.push(tpq.x);
              bfly.tbbody.push(tpq.y);
              bfly.obbody.push(0);
              bfly.obbody.push(0);
              const len = Math.sqrt(pqd.x * pqd.x + pqd.y * pqd.y);
              const ycomponent = (len > 0.0) ? (pqd.y / len) : 0.0;
              const n = new Vec3(Math.sin(a), ycomponent, Math.sin(a)).normalized();
              bfly.nbbody.push(n.f[1]);
              bfly.nbbody.push(n.f[2]);
              bfly.nbbody.push(n.f[0]);
            }
          }
        }
      }

      makeDataBuffers(gl, patch, indexOfData) {
        const jstart = Math.floor(patch * this.N / 8);
        const jend = (patch + 1) * Math.floor(this.boids.N / this.noBF / 8);
/*
        const offset = (this.dc * (this.c - 1));
        const loffset = (this.dc * (this.lc - 1));
        const totaloffset = offset * 2 + loffset * 2;
        const indexcount = totaloffset * 2 + 8;
        */
        // const ib = new Uint32Array(indexcount * this.boids.N);
        const bfly = this.bflies[indexOfData];

        for (let j = jstart; j < jend; j++) {
          const offsetbase = bfly.offsetbase;
          const ibl = [];
          const ibr = [];
          const libl = [];
          const libr = [];
          const ibbody = [];
          const ibla = [];
          const ibra = [];
          ibl.push(offsetbase);
          ibr.push(offsetbase + bfly.offset);
          for (let i = 0; i < bfly.offset; i++) {
            ibl.push(offsetbase + i);
            ibl.push(offsetbase);
            ibr.push(offsetbase + i + bfly.offset);
            ibr.push(offsetbase + bfly.offset);
          }
          ibl.push(offsetbase);
          ibr.push(offsetbase + bfly.offset);

          libl.push(offsetbase + 2 * bfly.offset);
          libr.push(offsetbase + 2 * bfly.offset + bfly.loffset);
          for (let i = 0; i < bfly.loffset; i++) {
            libl.push(offsetbase + 2 * bfly.offset + i);
            libl.push(offsetbase + 2 * bfly.offset);
            libr.push(offsetbase + 2 * bfly.offset + i + bfly.loffset);
            libr.push(offsetbase + 2 * bfly.offset + bfly.loffset);
          }
          libl.push(offsetbase + 2 * bfly.offset);
          libr.push(offsetbase + 2 * bfly.offset + bfly.loffset);

          const offsetbody = offsetbase + bfly.offset * 2 + bfly.loffset * 2;
          ibbody.push(offsetbody);
          for (let i = 0; i < bfly.cbody - 1; i++) {
            for (let k = 0; k < bfly.dcbody - 1; k++) {
              for (let n = 0; n < bfly.drbody; n++) {
                ibbody.push(offsetbody + i * bfly.dcbody * bfly.drbody + k * bfly.drbody + n);
                ibbody.push(offsetbody + i * bfly.dcbody * bfly.drbody + (k + 1) * bfly.drbody + n);
              }
            }
          }
          ibbody.push(ibbody[ibbody.length - 1]);

          const offsetantennae = offsetbase + bfly.offset * 2 + bfly.loffset * 2 + bfly.boffset;
          ibla.push(offsetantennae);
          ibra.push(offsetantennae + bfly.antennaeoffset);
          for (let i = 0; i < (bfly.ca - 1); i++) {
            for (let k = 0; k < (bfly.dca - 1); k++) {
              for (let n = 0; n < bfly.dra; n++) {
                ibla.push(offsetantennae + (i * bfly.dca * bfly.dra) + k * bfly.dra + n);
                ibla.push(offsetantennae + (i * bfly.dca * bfly.dra) + (k + 1) * bfly.dra + n);
                ibra.push(offsetantennae + bfly.antennaeoffset + (i * bfly.dca * bfly.dra) + k * bfly.dra + n);
                ibra.push(offsetantennae + bfly.antennaeoffset + (i * bfly.dca * bfly.dra) + (k + 1) * bfly.dra + n);
              }
            }
          }
          ibla.push(ibla[ibla.length - 1]);
          ibra.push(ibra[ibra.length - 1]);

          const ibt = ibl.concat(ibr).concat(libl).concat(libr).concat(ibbody).concat(ibla).concat(ibra);

          bfly.ib.set(ibt, bfly.indexcount * j);
          // ib = ib.concat(ibt);
          bfly.offsetbase += bfly.totaloffset;
        }
      }

      createGLBuffers(gl, indexOfData) {
        const bfly = this.bflies[indexOfData];
        const vb = new Float32Array(bfly.vbt.length * this.N);
        const tb = new Float32Array(bfly.tbt.length * this.N);
        const nb = new Float32Array(bfly.nbt.length * this.N);
        const ob = new Float32Array(bfly.obt.length * this.N);
        const inb = new Float32Array(bfly.vbt.length / 3 * this.N);
        for (let i = 0; i < this.N; i++) {
          vb.set(bfly.vbt, i * bfly.vbt.length);
          // vb = vb.concat(this.vbt);
          tb.set(bfly.tbt, i * bfly.tbt.length);
          // tb = tb.concat(this.tbt);
          ob.set(bfly.obt, i * bfly.obt.length);
          // ob = ob.concat(this.obt);
          nb.set(bfly.nbt, i * bfly.nbt.length);
          // nb = nb.concat(this.nbt);
          for (let j = 0; j < bfly.vbt.length / 3; j++) {
            inb[i * bfly.vbt.length / 3 + j] = i + indexOfData * this.N;
          }
        }
        bfly.verticeCount = vb.length / 3;
        bfly.vbuffer = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, bfly.vbuffer);
        gl.bufferData(gl.ARRAY_BUFFER, vb, gl.STATIC_DRAW);
        bfly.tbuffer = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, bfly.tbuffer);
        gl.bufferData(gl.ARRAY_BUFFER, tb, gl.STATIC_DRAW);
        bfly.obuffer = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, bfly.obuffer);
        gl.bufferData(gl.ARRAY_BUFFER, ob, gl.STATIC_DRAW);
        bfly.inbuffer = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, bfly.inbuffer);
        gl.bufferData(gl.ARRAY_BUFFER, inb, gl.STATIC_DRAW);
        bfly.nbuffer = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, bfly.nbuffer);
        gl.bufferData(gl.ARRAY_BUFFER, nb, gl.STATIC_DRAW);

        bfly.ibuffer = gl.createBuffer();
        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, bfly.ibuffer);
        gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, bfly.ib, gl.STATIC_DRAW);
        bfly.pc = bfly.ib.length;
      }
    }

    class Swarm {
      constructor(gl, canvas, data, position) {
        this.gl = gl;
        this.dataset = data;
        this.preSteps = 0;
        this.stage = 0;
        this.cb = null;
        this.position = position;
      }
      init(gl, canvas, textElem, callback) {
        if (this.cb === null) {
          this.cb = callback;
        }
        if (this.stage === 0) {
          this.stage++;
          this.swarm = new AcceleratedBoidsLib.Boids(gl, canvas, 2048, this.position);
          this.swarm.init(gl, canvas, textElem, callback);
          textElem.innerHTML += '<p>Creating Butterfly Renderer</p>';
          setTimeout(this.init(gl, canvas, textElem), 10);
          return;
        }
        if (this.stage === 1) {
          this.stage++;
          this.swarmRenderer = new ButterflyRenderer(this.swarm, gl, canvas, this.dataset);
          setTimeout(this.init(gl, canvas, textElem), 10);
          return;
        }
        if (this.stage === 2) {
          this.stage++;
          this.swarmRenderer.init(gl, canvas, textElem, () => this.init(gl, canvas, textElem));
          return;
        }
        if (this.stage === 3) {
          this.stage++;
          textElem.innerHTML += '<p>Stepping Simulation</p>';
          textElem.innerHTML += `<p id="stepid">Step ${this.preSteps}</p>`;
          setTimeout(this.init(gl, canvas, textElem), 10);
          return;
        }
        if (this.preSteps < 10000) {
          for (let i = 0; i < 1000; i++) {
            this.swarm.tick(gl, canvas, 0);
            this.swarm.runPhysics(gl, canvas, 0);
            this.preSteps++;
            document.getElementById('stepid').innerHTML = `Step ${this.preSteps}`;
          }
          setTimeout(this.init(gl, canvas, textElem), 10);
          return;
        }
        // this.debugDraw = new DebugLib.DebugTexture(gl);
        // setTimeout(() => callback(gl, canvas, textElem, callback), 10);
        this.cb();
      }

      tick(gl, canvas, t) {
        this.swarm.tick(gl, canvas, t);
        this.swarmRenderer.tick(gl, canvas, t);
      }

      runPhysics(t) {
        this.swarm.runPhysics(t);
      }

      setColorAdd(c) {
        this.swarmRenderer.colorAdd = c;
      }

      setSpeed(s) {
        this.swarmRenderer.scrollSpeed = s;
      }

      draw(gl, canvas, pm, vm, cm, mm) {
        this.swarmRenderer.draw(gl, canvas, pm, vm, cm, mm);
/*
        this.debugDraw.setMultiplier([0.0001, 2.0, 0.1]);
        this.debugDraw.setOffset([0.0, 0.0, 0.0]);
        this.debugDraw.draw(gl, this.swarm.nfb, 2048, 384, 0, 0);


        this.debugDraw.setMultiplier([100.0, 100.0, 100.0]);
        this.debugDraw.setOffset([0.5, 0.5, 0.5]);
        this.debugDraw.draw(gl, this.swarm.cfb, this.swarm.N, 8, 0, 0);
        this.debugDraw.setMultiplier([100.0, 100.0, 100.0]);
        this.debugDraw.setOffset([0.5, 0.5, 0.5]);
        this.debugDraw.draw(gl, this.swarm.ofb, this.swarm.N, 8, 0, 32);
        */
      }
    }

    SwarmLib.Swarm = Swarm;
    SwarmLib.ButterflyRenderer = ButterflyRenderer;

    return SwarmLib;
  }
// define globally if it doesn't already exist
  if (typeof (SwarmLib) === 'undefined') {
    const mwindow = window;
    mwindow.SwarmLib = defineLibrary();
  }
}(window)); // eslint-disable-line no-undef
