(function capsule(window) {
  function defineLibrary() {
    const GastropodaLib = {};
    const Vec3 = VemaLib.Vec3;
    const Matrix4 = VemaLib.Matrix4;

    GastropodaLib.generator = new Random.Generator(12433,46831,66713,11447);

    class Shell {
      constructor(gl, canvas, size, position) {
        this.position = position;
        this.size = size;
        this.colorCount= 8;
        this.changingColors = [37/255,24/255,15/255,1.0,
                              81/255,53/255,31/255,1.0,
                              159/255,133/255,110/255,1.0,
                              134/255,77/255,67/255,1.0,
                              104/255,67/255,61/255,1.0,
                              118/255,75/255,56/255,1.0,
                              187/255,158/255,98/255,1.0,
                              159/255,111/255,65/255,1.0];
        this.colorAdd = [0.0,0.0,0.0];
      }

      setRGB(color){
        this.colorAdd =color;
      };

      prerender(gl, canvas, tree, vData, cData, nData, iData) {
        const subdivisionA = 3;
        const subdivisionB = 300;
        const subdivisionC = 600;
        const spiralA = 1;
        const spiralB = 0.2;
        const axisA = new Vec3(0, 0, 1).normalized();
        const axisB = new Vec3(1, 0, 0).normalized();
        const axisC = new Vec3(0, 1, 0).normalized();
        const axisD = axisA;//new Vec3(-0.4,0.4,0.7).normalized();
        const lemmy =1;
        for (let i = 0; i < subdivisionA * subdivisionB; i++) {
          const t = i * Math.PI * 2.0 / subdivisionB;
          const t0 = Math.max(0, t - Math.PI * 2);
          const l0 = spiralA *(Math.E**(spiralB * t0));
          const rot0 = new Matrix4();
          rot0.rotateMatrix(t0, axisD);
          const trans0 = new Matrix4();
          const tr0 = axisA.multiple(l0*lemmy);
          trans0.translate(-tr0.f[0],-tr0.f[1],-tr0.f[2]);
          const c0 = (trans0.multiple(rot0)).multipleV(axisB.multiple(l0));
          const l = spiralA * (Math.E**(spiralB * t));
          const rot = new Matrix4();
          rot.rotateMatrix(t, axisD);
          const trans = new Matrix4();
          const tr = axisA.multiple(l*lemmy);
          trans.translate(-tr.f[0],-tr.f[1],-tr.f[2]);
          const c = (trans.multiple(rot)).multipleV(axisB.multiple(l));
          const t1 = t + Math.PI * 2;
          const l1 = spiralA * (Math.E**(spiralB * t1));
          const rot1 = new Matrix4();
          rot1.rotateMatrix(t1, axisD);
          const trans1 = new Matrix4();
          const tr1 = axisA.multiple(l1*lemmy);
          trans1.translate(-tr1.f[0],-tr1.f[1],-tr1.f[2]);
          const c1 = (trans1.multiple(rot1)).multipleV(axisB.multiple(l1));
          const indexi = i%this.colorCount;
          const cm = 0.3;
          let rough =0.005*GastropodaLib.generator.random();
          for (let j = 0; j < subdivisionC; j++) {
            const fade =GastropodaLib.generator.random()*cm;
            const color = [this.changingColors[indexi*4]-fade,
            this.changingColors[indexi*4+1]-fade,
            this.changingColors[indexi*4+2]-fade,
            this.changingColors[indexi*4+3]];
            const n = c.normalized();
            const angle = Math.PI * 2 * j / (subdivisionC - 1);
            const rot = new Matrix4();

            const r1 = c1.subtract(c).multiple(0.5);
            const r0 = c0.subtract(c).multiple(0.5);
            rot.rotateMatrix(t, axisA);
            //const cosa = Math.cos(angle);
            const roty = new Matrix4();
            roty.rotateMatrix(angle, axisC);
            const ra = roty.multipleV(axisB);
            //const d = (cosa + 1) / 2;
            const d1 = ra.normalized().dot(r1.normalized());
            const d0 = ra.normalized().dot(r0.normalized());
            //const d = d1;
            if(j%2===0){
              rough =0.005*GastropodaLib.generator.random();
            }

            const d11 = (d1+1)/2*(1+0.01*Math.sin(i)-rough);
            const d00 = (d0+1)/2*(1+0.01*Math.sin(i)-rough);
            const ca = c.add(rot.multipleV(ra.multiple(r1.length * d11 + r0.length * d00)));
            const a = this.position.add(ca.multiple(this.size));

            const normal = ra.normalized();
            if(j <subdivisionC-1){
              vData.push(a.f[0]);
              vData.push(a.f[1]);
              vData.push(a.f[2]);
              nData.push(normal.f[0]);
              nData.push(normal.f[1]);
              nData.push(normal.f[2]);
              cData.push(color[0]);
              cData.push(color[1]);
              cData.push(color[2]);
              cData.push(1.0);
            }else{
              const vx =vData[vData.length-(subdivisionC-1)*3];
              const vy =vData[vData.length-(subdivisionC-1)*3+1];
              const vz =vData[vData.length-(subdivisionC-1)*3+2];
              const nx = nData[nData.length-(subdivisionC-1)*3];
              const ny = nData[nData.length-(subdivisionC-1)*3+1];
              const nz = nData[nData.length-(subdivisionC-1)*3+2];
              const cx = cData[cData.length-(subdivisionC-1)*4];
              const cy = cData[cData.length-(subdivisionC-1)*4+1];
              const cz = cData[cData.length-(subdivisionC-1)*4+2];
              vData.push(vx);
              vData.push(vy);
              vData.push(vz);
              nData.push(nx);
              nData.push(ny);
              nData.push(nz);
              cData.push(cx);
              cData.push(cy);
              cData.push(cz);
              cData.push(1.0);
            }

            if (i < subdivisionA * subdivisionB - 1) {
              iData.push(i * subdivisionC + j);
              iData.push(i * subdivisionC + subdivisionC + j);
            }
          }
        }
        this.vData = vData;
        this.cData = cData;
        this.nData = nData;
        this.iData = iData;
        this.calculateNormals(subdivisionA,subdivisionB,subdivisionC);
        this.len = iData.length;
        console.log(`vertices ${this.len}`);
      }

      calculateNormals(subdivisionA,subdivisionB, subdivisionC){
        for (let i = 0; i < subdivisionA * subdivisionB; i++) {
          for (let j = 0; j < subdivisionC; j++) {
            const o= i*subdivisionC+j;
            const i0 = o;
            const i1 = i*subdivisionC+(j+subdivisionC-1)%subdivisionC;
            const i2 = i*subdivisionC+(j+1)%subdivisionC;
            const p0 = new Vec3(this.vData[3*i0],this.vData[3*i0+1],this.vData[3*i0+2]);
            const p1 = new Vec3(this.vData[3*i1],this.vData[3*i1+1],this.vData[3*i1+2]);
            const p2 = new Vec3(this.vData[3*i2],this.vData[3*i2+1],this.vData[3*i2+2]);
            let count =0;
            const sum = new Vec3(0,0,0);
            if(i > 0){
              const i3 =i*subdivisionC - subdivisionC+(j+subdivisionC-1)%subdivisionC;
              const i4 =o - subdivisionC;
              const p3 = new Vec3(this.vData[3*i3],this.vData[3*i3+1],this.vData[3*i3+2]);
              const p4 = new Vec3(this.vData[3*i4],this.vData[3*i4+1],this.vData[3*i4+2]);

              sum.addToThis((p0.subtract(p1)).cross(p3.subtract(p0)).normalized());
              sum.addToThis((p4.subtract(p0)).cross(p3.subtract(p4)).normalized());
              sum.addToThis((p2.subtract(p0)).cross(p4.subtract(p2)).normalized());
            }
            if (i < subdivisionA * subdivisionB - 1) {
              const i5 = o + subdivisionC;
              const i6 = i*subdivisionC+ subdivisionC+ (j+subdivisionC+1)%subdivisionC;

              const p5 = new Vec3(this.vData[3*i5],this.vData[3*i5+1],this.vData[3*i5+2]);
              const p6 = new Vec3(this.vData[3*i6],this.vData[3*i6+1],this.vData[3*i6+2]);
              sum.addToThis((p5.subtract(p1)).cross(p0.subtract(p5)).normalized());
              sum.addToThis((p5.subtract(p0)).cross(p6.subtract(p5)).normalized());
              sum.addToThis((p6.subtract(p0)).cross(p2.subtract(p6)).normalized());
            }

            const normal = sum.normalized();

            this.nData[o*3] = normal.f[0];
            this.nData[o*3+1] = normal.f[1];
            this.nData[o*3+2] = normal.f[2];

          }
        }
      }

      render(gl, canva, program, params) {
        gl.useProgram(program);
        this.vertices = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, this.vertices);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(this.vData), gl.STATIC_DRAW);
        this.colors = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, this.colors);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(this.cData), gl.STATIC_DRAW);
        this.normals = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, this.normals);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(this.nData), gl.STATIC_DRAW);
        this.indices = gl.createBuffer();
        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.indices);
        gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, new Uint32Array(this.iData), gl.STATIC_DRAW);
      }

      draw(gl, canva, program, params) {
        gl.uniform3f(params.colorAdd,this.colorAdd[0],this.colorAdd[1],this.colorAdd[2]);
        gl.bindBuffer(gl.ARRAY_BUFFER, this.vertices);
        gl.enableVertexAttribArray(params.aVertexPosition);
        gl.vertexAttribPointer(params.aVertexPosition, 3, gl.FLOAT, false, 0, 0);
        gl.bindBuffer(gl.ARRAY_BUFFER, this.normals);
        gl.enableVertexAttribArray(params.aNormalPosition);
        gl.vertexAttribPointer(params.aNormalPosition, 3, gl.FLOAT, false, 0, 0);
        gl.bindBuffer(gl.ARRAY_BUFFER, this.colors);
        gl.enableVertexAttribArray(params.uColor);
        gl.vertexAttribPointer(params.uColor, 4, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.indices);
        gl.drawElements(gl.TRIANGLE_STRIP, this.len, gl.UNSIGNED_INT, 0);
        // gl.drawArrays(gl.TRIANGLE_STRIP, 0, this.len);
      }
    }


    class Gastropoda {
      constructor(gl, canvas) {
        this.t = 0;
        this.vb = `#version 300 es
    in vec3 inputPosition;
    in vec3 inputNormal;
    in vec4 uColor;
    	uniform mat4 projection, modelview, uL, uC;
    out vec4 color;
    out vec3 u;
    out vec3 n;
    	void main(){

        color=uColor;
        u=(uC*modelview*vec4(inputPosition,1.)).xyz;
        n=normalize((uC*modelview*vec4(inputNormal,0.)).xyz);
        gl_Position = projection*uC*uL * modelview * vec4(inputPosition,1.);
    	}
    `;

        this.fb =
    `#version 300 es
      precision mediump float;
    	in vec4 color;
      in vec3 u;
      in vec3 n;
      uniform vec3 k;
      uniform vec3 colorAdd;
      out vec4 fragColor;
      const float screenGamma =2.2;

      void main() {
         vec4 texCol = vec4(1.0,1.0,1.0,1.0);
         vec4 c = vec4(color.rgb+colorAdd,1.0);

          vec3 ambientColor = c.xyz/4.0;
        vec3 diffuseColor = c.xyz;
          vec3 normal = normalize(n);
        vec3 lightDir = normalize(k-u);

          float lambertian = max(abs(dot(normal,lightDir)), 0.0);

          vec4 colorLinear = vec4(ambientColor +
                             lambertian * diffuseColor,1.);
          fragColor = colorLinear;
          //vec4 colorGammaCorrected = vec4(pow(colorLinear, vec3(1.0/screenGamma)),1.0);
          //float vd=clamp((-20.+length(k-u))/20.,0.,1.);
          //gl_FragColor=mix(colorLinear,vec4(.9,.9,.9,1.),vd);
        }`;

        this.program = ShaderLib.makeProgram(gl, this.vb, this.fb);
        gl.useProgram(this.program);
        this.params = {};
        this.params.projection = gl.getUniformLocation(this.program, 'projection');
        this.params.modelview = gl.getUniformLocation(this.program, 'modelview');
        this.params.uC = gl.getUniformLocation(this.program, 'uC');
        this.params.uL = gl.getUniformLocation(this.program, 'uL');
        this.params.k = gl.getUniformLocation(this.program, 'k');
        this.params.colorAdd = gl.getUniformLocation(this.program, 'colorAdd');
        this.params.uColor = gl.getAttribLocation(this.program, 'uColor');
        this.params.aVertexPosition = gl.getAttribLocation(this.program, 'inputPosition');
        this.params.aNormalPosition = gl.getAttribLocation(this.program, 'inputNormal');
        this.shell = new Shell(gl, canvas, 0.025, new Vec3(0.0, 0.0, 0.0));
        this.shellVertexBuffer = [];
        this.shellColorBuffer = [];
        this.shellNormalBuffer = [];
        this.shellIndexBuffer = [];
        this.shell.prerender(gl, canvas, this.tree, this.shellVertexBuffer, this.shellColorBuffer, this.shellNormalBuffer, this.shellIndexBuffer);
        this.shell.render(gl, canvas, this.program, this.params);
      }


      draw(gl, canvas) {
        const aspectRatio = canvas.width / canvas.height;
        const fov = 45;
        const viewangle = (35 * Math.PI) / 180;
        const viewdistance = 0.1 / Math.tan(((fov / 2) * Math.PI) / 180);
        this.perspectiveMatrix = new Matrix4();
        this.cameraMatrix = new Matrix4();
        this.model = new Matrix4();
        this.view = new Matrix4();
        this.modelPos = new Vec3(0.0, 0.0, 0.0);
        const roty = new Matrix4();
        const rotx = new Matrix4();
        rotx.rotationXMatrix(this.t*Math.PI/3);
        roty.rotationYMatrix(this.t / 10.0);
        const transy = new Matrix4();
        transy.translate(0, 0, 0);
        this.model = transy.multiple(rotx.multiple(roty));
        this.cameraPos = new Vec3(0, 5, 6);
        this.cameraUp = new Vec3(0, 1, 0);
        this.cameraFront = new Vec3(0, 0, 1);
        this.cameraDirection = 0.0;
        this.light = new Float32Array([5.0, 40.0, -3]);
        this.dirMatrix = new Matrix4();
        this.dirMatrix.rotateMatrix(-this.cameraDirection, this.cameraUp);
        this.cameraFront = this.dirMatrix.multipleV(this.cameraFront);
        const rotateAxis = this.cameraFront.cross(this.cameraUp).normalized();
        const rot2 = new Matrix4();
        rot2.rotateMatrix(-viewangle, rotateAxis);
        this.cameraNormal = rot2.multipleV(this.cameraFront);
        this.cameraUp = rot2.multipleV(this.cameraUp);

        this.cameraMatrix.makeLookAt(this.cameraNormal, this.cameraUp, this.cameraUp.cross(this.cameraNormal).normalized(), this.cameraPos);
        this.perspectiveMatrix.makePerspective(fov, aspectRatio, viewdistance, 100.0);

        gl.useProgram(this.program);
        gl.clearColor(0.3, 0.6, 0.95, 1.0);
        // Enable depth testing
        gl.enable(gl.DEPTH_TEST);
        // Near things obscure far things
        gl.depthFunc(gl.LEQUAL);
        // Clear the color as well as the depth buffer.
        gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT); // eslint-disable-line no-bitwise
        gl.viewport(0, 0, canvas.width, canvas.height);

        gl.uniformMatrix4fv(this.params.projection, false, new Float32Array(this.perspectiveMatrix.m));
        gl.uniformMatrix4fv(this.params.modelview, false, new Float32Array(this.model.m));
        gl.uniformMatrix4fv(this.params.uC, false, new Float32Array(this.cameraMatrix.m));
        gl.uniformMatrix4fv(this.params.uL, false, new Float32Array(this.view.m));
        gl.uniform3fv(this.params.k, this.light);
        this.shell.draw(gl, canvas, this.program, this.params);
      }

      setT(t) {
        this.t = t;
      }

      setRGB(color){
        this.shell.setRGB(color);
      }

    }

    GastropodaLib.Gastropoda = Gastropoda;

    return GastropodaLib;
  }
// define globally if it doesn't already exist
  if (typeof (GastropodaLib) === 'undefined') {
    const mwindow = window;
    mwindow.GastropodaLib = defineLibrary();
  }
}(window)); // eslint-disable-line no-undef
