(function capsule(window) {
  function defineLibrary() {
    const Vec3 = VemaLib.Vec3; // eslint-disable-line no-undef
    const GrassLib = {};

    class Grass {
      constructor(gl, canvas) {
        this.gvbuffer = null;
        this.calcsegmentbuffer = null;
        this.gibuffer = null;
        this.vbuffer = null;
        this.tbuffer = null;
        this.gi2buffer = null;
        this.grassprogram = null;

        this.calctexture = null;
        this.calcframebuffer = null;

        this.calctexture0 = null;
        this.calcframebuffer0 = null;
        this.calctexture1 = null;
        this.calcframebuffer1 = null;

        this.calculateProgram = null;

        this.floorheight = -1.7;
        this.wallx = 1.0;
        this.wallz = 1.0;

        this.patchesxstart = 0;
        this.patchesystart = 0;
        this.patchesxend = 1;
        this.patchesyend = 1;

        this.w = 400;
        this.h = 400;

        this.phase = 0.0;

        this.buffernumber = 0;

        this.mRadii = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
          0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0];
        this.mMidPoints = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
          0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
          0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
          0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
          0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
          0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0];
        this.mObjectCount = 0;

        this.segments = 5;
        this.ss = this.segments - 1;

        this.direction = Math.random();
        this.windDir = [0.0, 0.0, 0.0, 0.6];

        this.createBuffers(gl);
        this.setGrass(40.0, 40.0, GroundLib.ground(0.0, 0.0));
        this.makeGrassBuffers(gl);
      }

      setGrass(x, z, h) {
        this.floorheight = h;
        this.wallx = x;
        this.wallz = z;
      }

      setObjects(midPoints, radii) {
        for (let i = 0; i < 32; i++) {
          const os = i * 3;
          if (i < radii.length) {
            mMidPoints[os] = midPoints[os];
            mMidPoints[os + 1] = midPoints[os + 1];
            mMidPoints[os + 2] = midPoints[os + 2];
            mRadii[i] = radii[i];
          } else {
            mMidPoints[os] = 0;
            mMidPoints[os + 1] = 0;
            mMidPoints[os + 2] = 0;
            mRadii[i] = 0;
          }
        }
        objectCount = radii.length;
      }

      makeGrassBuffers(gl) {
        gl.bindBuffer(gl.ARRAY_BUFFER, this.gvbuffer);
        const varray = new Float32Array(this.w * this.h * 2 * 4 * this.segments);
        let offset = 0;
        for (let i = 0; i < this.w; i++) {
          for (let j = 0; j < this.h; j++) {
            // const rot = Math.random();
            const len = Math.random();
            const xpos = -this.wallx / 2.0 + 1.0 * this.wallx * (i + Math.random() - 0.5) / this.w;
            const ypos = +this.wallz / 2.0 - 1.0 * this.wallz * (j + Math.random() - 0.5) / this.h;
            const zpos = GroundLib.ground(xpos, ypos);
            for (let k = 0; k < 2 * this.segments; k++) {
              varray[offset++] = xpos;
              varray[offset++] = ypos;
              varray[offset++] = zpos;
              // varray[offset++] = rot;
              varray[offset++] = len;
            }
          }
        }

        gl.bufferData(gl.ARRAY_BUFFER,
        varray, gl.STATIC_DRAW);

        gl.bindBuffer(gl.ARRAY_BUFFER, this.calcsegmentbuffer);

        const calcsegmentarray = new Float32Array(this.w * this.h * 2 * 4 * this.segments);
        const sp = 1.0;
        offset = 0;
        for (let i = 0; i < this.w; i++) {
          for (let j = 0; j < this.h; j++) {
            for (let k = 0; k < this.segments; k++) {
              calcsegmentarray[offset++] = k;
              calcsegmentarray[offset++] = i;
              calcsegmentarray[offset++] = j;
              calcsegmentarray[offset++] = sp;
              calcsegmentarray[offset++] = k;
              calcsegmentarray[offset++] = i;
              calcsegmentarray[offset++] = j;
              calcsegmentarray[offset++] = -sp;
            }
          }
        }

        gl.bufferData(gl.ARRAY_BUFFER,
          calcsegmentarray, gl.STATIC_DRAW);
      }
      initBuffers(gl) {
        this.vbuffer = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, this.vbuffer);
        const varray = new Float32Array(this.w * this.h * 2 * 2 * this.segments);
        let offset = 0;
        for (let i = 0; i < this.w; i++) {
          for (let j = 0; j < this.h; j++) {
            for (let k = 0; k < this.segments; k++) {
              varray[offset++] = -1.0 + (i + 0.5) * 2.0 / this.w;
              varray[offset++] = -1.0 + ((j * 2 + (k <= 2.0 ? 0.0 : 1.0)) / 2 + 0.25) * 2.0 / this.h;

              varray[offset++] = -1.0 + (i + 0.5) * 2.0 / this.w;
              varray[offset++] = -1.0 + ((j * 2 + (k <= 2.0 ? 0.0 : 1.0)) / 2 + 0.25) * 2.0 / this.h;
            }
          }
        }
        gl.bufferData(gl.ARRAY_BUFFER, varray, gl.STATIC_DRAW);
        this.tbuffer = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, this.tbuffer);
        const tarray = new Float32Array(this.w * this.h * 2 * 2 * this.segments);
        offset = 0;
        for (let i = 0; i < this.w; i++) {
          for (let j = 0; j < this.h; j++) {
            for (let k = 0; k < 2 * this.segments; k++) {
              tarray[offset++] = (i + 0.5) * 1.0 / this.w;
              tarray[offset++] = (j + 0.5) * 1.0 / this.h;
            }
          }
        }

        gl.bufferData(gl.ARRAY_BUFFER,
        tarray, gl.STATIC_DRAW);
      }

      createBuffers(gl) {
        const grassvertex = `attribute vec4 inputPosition;
      attribute vec4 calcSegment;

      uniform ivec2 uSize;
      uniform mat4 projection, model, view, camera;
      uniform sampler2D uSampler;
      uniform float uFloorHeight;
      uniform vec3 uOffset;
      varying vec3 normalInterp;
      varying vec3 vertPos;
      varying vec3 vColor;
      void main(){
        vec4 calc0 = texture2D(uSampler,vec2(1.0/float(uSize.x),1.0/(float(uSize.y)*2.0))*(vec2(0.5,0.5)+vec2(calcSegment.y,2.0*calcSegment.z)));
        vec4 calc1 = texture2D(uSampler,vec2(1.0/float(uSize.x),1.0/(float(uSize.y)*2.0))*(vec2(0.5,0.5)+vec2(calcSegment.y,2.0*calcSegment.z+1.0)));
          vec3 stem = vec3(0.0,(0.5+inputPosition.w)/16.0,0.0);

          float ml = 2.0;
          float x= (inputPosition.x+uOffset.x)*ml;
          float y= (inputPosition.y+uOffset.z)*ml;
          float level =0.0;//1.6*(0.04*sin(x*2.78)+0.16*sin(x*1.58)-0.14*sin(y*0.86)+0.06*sin(y*3.58)+0.1*cos((x+y)*1.34));

          float ry = 3.14159*2.0*(inputPosition.w-0.5);
          float rx2 = 3.14159*0.5*(inputPosition.w);
          mat3 roty;
          roty[0] = vec3(cos(ry),-sin(ry),0.0);
          roty[1] = vec3(-sin(ry),cos(ry),0.0);
          roty[2] = vec3(0.0,0.0,1.0);

          mat3 rotx2;
          rotx2[0] = vec3(1.0,0.0,0.0);
          rotx2[1] = vec3(0.0,cos(rx2),-sin(rx2));
          rotx2[2] = vec3(0.0,-sin(rx2),cos(rx2));


          vec3 iPosition = vec3(inputPosition.x,inputPosition.z,inputPosition.y);
          float width = (5.0-(calcSegment.x+1.0))/5.0;
          width=sqrt(width);
          vec3 rotation = vec3(calcSegment.w*0.015*width,0.0,0.0);
          vec3 normroot = vec3(1.0,0.0,0.0);
          rotation = roty*rotation;
          stem = rotx2*stem;

          vec3 root = iPosition+rotation+uOffset;
          vec3 rotatedStem = vec3(0.0);

          float rx = 0.0;
          float rz = 0.0;

          for(int i=0; i< 4; i++){
            bool replace = calcSegment.x>float(i)+0.1;
            rx = i<1?(calc0.x-0.5):i<2?(calc0.z-0.5):i<3?(calc1.x-0.5):(calc1.z-0.5);
            rz = i<1?(calc0.y-0.5):i<2?(calc0.w-0.5):i<3?(calc1.y-0.5):(calc1.w-0.5);
            mat3 rotx;
            rotx[0] = vec3(1.0,0.0,0.0);
            rotx[1] = vec3(0.0,cos(rx),-sin(rx));
            rotx[2] = vec3(0.0,-sin(rx),cos(rx));
            mat3 rotz;
            rotz[0] = vec3(cos(rz),0.0,-sin(rz));
            rotz[1] = vec3(0.0,1.0,0.0);
            rotz[2] = vec3(sin(rz),0.0,cos(rz));

            rotatedStem = rotz*rotx*stem;
            stem = replace?rotatedStem:stem;
            root = replace?(root+rotatedStem):root;

          }

          vColor = mix(vec3(0.5,0.8,0.5),vec3(calc1.rgb),calcSegment.x/3.0); //:vec3(calc1.bgr);//(calcSegment.x>0.5)?vec3(calc1.rgb):vec3(calc1.bgr);//(calcSegment.x>0.5)?vec3(rx+0.5,rz+0.5,0.5):vec3(0.5,rz+0.5,rx+0.5);////
          //vec3(0.5,0.8,0.5);// vec3(0.1,0.9,0.1);//

          vec3 finalPosition = root;//(calcSegment.x>0.5)?(root+rotatedStem):root;
          vec3 normal=cross(rotatedStem,cross(roty*normroot,rotatedStem));
          vertPos = finalPosition;
          normalInterp = normal;
          gl_Position = projection * camera * view * model * vec4(finalPosition, 1.0);

      }`;
        const grassfragment = `precision mediump float;
      const float shininess = 16.0;
    	const float screenGamma = 2.2; // Assume the monitor is calibrated to the sRGB color space
      varying vec3 normalInterp;
      varying vec3 vertPos;
      varying vec3 vColor;
      void main(){
        //gl_FragColor =vec4(vColor,1.0); //vec4(0.4,0.8,0.4,1.0);

      vec4 c1 =vec4(vColor,1.0);
      float luma = 0.0;
      luma =0.2;
      vec3 uLightPos = vec3(10.0,3000.0,100.0);
      float lumambient = luma/2.0;
      //vec4 c = vec4(mix(uColor.xyz,uColor3.xyz,luma),1.0);
      //vec4 c = vec4(vColor,1.0);
      vec4 c = vec4(0.05,0.2,0.05,1.0);
  		//c.a = c1.a;
  	   vec3 ambientColor = c.xyz/4.0;
  	   vec3 diffuseColor = c.xyz;

  	  vec3 specColor = vec3(0.8,0.8,0.8);
  	  vec3 normal = normalize(normalInterp);
  	  vec3 lightDir = normalize(uLightPos - vertPos);

  	  float lambertian = max(dot(lightDir,normal), 0.0);
  	  float specular = 0.0;

  	  if(lambertian > 0.0) {

  	    vec3 viewDir = normalize(-vertPos);
  	    // this is blinn phong
  	    vec3 halfDir = normalize(lightDir + viewDir);
  	    float specAngle = max(dot(halfDir, normal), 0.0);
  	    specular = pow(specAngle, shininess);
  		vec3 reflectDir = reflect(-lightDir, normal);
  		      specAngle = max(dot(reflectDir, viewDir), 0.0);
  		      // note that the exponent is different here
  		      specular = pow(specAngle, shininess);
  	  }
  	  vec3 colorLinear = ambientColor +
  	                     lambertian * diffuseColor +
  	                     specular * specColor;
  	  // apply gamma correction (assume ambientColor, diffuseColor and specColor
  	  // have been linearized, i.e. have no gamma correction in them)
  	  vec3 colorGammaCorrected = pow(colorLinear, vec3(1.0/screenGamma));
  	  // use the gamma corrected color in the fragment
  	  gl_FragColor = vec4(colorGammaCorrected, 1.0);
    }`;


        this.grassprogram = ShaderLib.makeProgram(gl, grassvertex, grassfragment);
        this.gvbuffer = gl.createBuffer();
        this.calcsegmentbuffer = gl.createBuffer();
        this.makeGrassBuffers(gl);

        this.gibuffer = gl.createBuffer();
        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.gibuffer);
        const iarray = new Uint32Array(this.w * this.h * 2 * (this.segments + 2));
        let offset = 0;
        for (let i = 0; i < this.w * this.h; i++) {
          for (let j = 0; j < this.segments + 2; j++) {
            if (j == 0) {
              iarray[offset++] = i * 2 * this.segments;
              iarray[offset++] = i * 2 * this.segments;
            } else if (j == this.segments + 1) {
              iarray[offset++] = i * 2 * this.segments + 1 + (j - 2) * 2;
              iarray[offset++] = i * 2 * this.segments + 1 + (j - 2) * 2;
            } else {
              iarray[offset++] = i * 2 * this.segments + (j - 1) * 2;
              iarray[offset++] = i * 2 * this.segments + 1 + (j - 1) * 2;
            }
          }
        }
        gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, iarray, gl.STATIC_DRAW);

        this.gi2buffer = gl.createBuffer();
        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.gi2buffer);
        const iarray2 = new Uint32Array(this.w * this.h * 2);
        offset = 0;
        for (let i = 0; i < this.w * this.h; i++) {
          for (let j = 0; j < 2; j++) {
            iarray2[offset++] = i * 2 * this.segments + (j + 1) * 3;
          }
        }
        gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, iarray2, gl.STATIC_DRAW);
        this.calculateProgram = this.makeCalculateProgram(gl);
        this.initBuffers(gl);

        this.calctexture0 = gl.createTexture();
        this.calctextureunit0 = FrameBufferLib.getNextTextureUnit();
        this.calcframebuffer0 = FrameBufferLib.createDataFramebufferRGBA(gl, this.calctexture0, this.w, this.h);
        this.calctexture1 = gl.createTexture();
        this.calctextureunit1 = FrameBufferLib.getNextTextureUnit();
        this.calcframebuffer1 = FrameBufferLib.createDataFramebufferRGBA(gl, this.calctexture1, this.w, this.h);
        const can = document.createElement('canvas');
        can.width = this.w;
        can.height = this.h * 2;
        const ctx = can.getContext('2d');

        ctx.fillStyle = '#808080FF';
        ctx.fillRect(0, 0, this.w, this.h * this.ss);
        for (let j = 0; j < this.h; j++) {
          for (let i = 0; i < this.w; i++) {
            const gray = Math.floor(Math.random() * 255.0);
            ctx.fillStyle = `rgb(${gray},${gray},${gray})`;
          // ctx.fillRect(i, j, 1, 1);
          }
        }
        gl.bindTexture(gl.TEXTURE_2D, this.calctexture0);
        gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, can);
        gl.bindTexture(gl.TEXTURE_2D, this.calctexture1);
        gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, can);
        this.calctexture = this.calctexture0;
        this.calctextureunit = this.calctextureunit0;

        this.grassprogram.uPerspectiveMatrix = gl.getUniformLocation(this.grassprogram, 'projection');
        this.grassprogram.uCamera = gl.getUniformLocation(this.grassprogram, 'camera');
        this.grassprogram.uModel = gl.getUniformLocation(this.grassprogram, 'model');
        this.grassprogram.uView = gl.getUniformLocation(this.grassprogram, 'view');
        this.grassprogram.uSize = gl.getUniformLocation(this.grassprogram, 'uSize');

        this.grassprogram.uOffset = gl.getUniformLocation(this.grassprogram, 'uOffset');
      // grassprogram.uNormalMat = gl.getUniformLocation(grassprogram, 'normalMat');
        this.grassprogram.aVertexPosition = gl.getAttribLocation(this.grassprogram, 'inputPosition');

        this.grassprogram.aCalcSegmentPosition = gl.getAttribLocation(this.grassprogram, 'calcSegment');
      // grassgrassprogram.uLightPos = gl.getUniformLocation(program, 'uLightPos');
        this.grassprogram.uSampler = gl.getUniformLocation(this.grassprogram, 'uSampler');
        this.grassprogram.uFloorHeight = gl.getUniformLocation(this.grassprogram, 'uFloorHeight');

/*
        this.createGroundBuffers(gl);
        this.groundprogram = this.createGroundProgram(gl);
        this.initTextures(gl);
*/
        return this.grassprogram;
      }
      makeCalculateProgram(gl) {
        const calculatevertex = `
  precision highp float;
  attribute vec2 position;
  attribute vec2 texCoordinate;
  attribute vec4 inputPosition;
  attribute vec4 calcSegment;

  uniform ivec2 uSize;
  uniform float uFloorHeight;
  uniform sampler2D uTexture0;
  uniform vec4 uWindDir;

  varying vec2 v_TexCoordinate;
  varying vec3 vRoot;
  varying vec3 vPosition0;
  varying vec3 vPosition1;
  varying vec4 vColor;
  varying vec3 vWindForce;
  void main(){
      gl_PointSize = 1.0;

      vec4 calc0 = texture2D(uTexture0,vec2(1.0/float(uSize.x),1.0/float(uSize.y)*2.0)*(vec2(0.5,0.5)+vec2(calcSegment.y,2.0*calcSegment.z)));
      vec4 calc1 = texture2D(uTexture0,vec2(1.0/float(uSize.x),1.0/float(uSize.y)*2.0)*(vec2(0.5,0.5)+vec2(calcSegment.y,2.0*calcSegment.z+1.0)));
        vec3 stem = vec3(0.0,(0.5+inputPosition.w)/16.0,0.0);

        float ml = 0.8;
        float x= (inputPosition.x)*ml;
        float y= (inputPosition.y)*ml;
        float level =0.0;//1.6*(0.04*sin(x*2.78)+0.16*sin(x*1.58)-0.14*sin(y*0.86)+0.06*sin(y*3.58)+0.1*cos((x+y)*1.34));

        float ry = 3.14159*4.0*(inputPosition.w-0.5);
        float rx2 = 3.14159*0.5*(inputPosition.w);
        mat3 roty;
        roty[0] = vec3(cos(ry),-sin(ry),0.0);
        roty[1] = vec3(-sin(ry),cos(ry),0.0);
        roty[2] = vec3(0.0,0.0,1.0);

        mat3 rotx2;
        rotx2[0] = vec3(1.0,0.0,0.0);
        rotx2[1] = vec3(0.0,cos(rx2),-sin(rx2));
        rotx2[2] = vec3(0.0,-sin(rx2),cos(rx2));
        vec3 normRoot =roty*vec3(1.0,0.0,0.0);

        vec3 iPosition = vec3(inputPosition.x,inputPosition.z,inputPosition.y);
        float width = (5.0-(calcSegment.x+1.0))/5.0;
        width=sqrt(width);
        vec3 rotation = vec3(calcSegment.w*0.015*width,0.0,0.0);
      stem = rotx2*stem;

      vec3 root = iPosition+rotation;
      vec3 rotatedStem = stem;
      float rx = 0.0;
      float rz = 0.0;
      vRoot=root;
      for(int i=0; i< 4; i++){
        rx = i<1?(calc0.x-0.5):i<2?(calc0.z-0.5):i<3?(calc1.x-0.5):(calc1.z-0.5);
        rz = i<1?(calc0.y-0.5):i<2?(calc0.w-0.5):i<3?(calc1.y-0.5):(calc1.w-0.5);
        mat3 rotx;
        rotx[0] = vec3(1.0,0.0,0.0);
        rotx[1] = vec3(0.0,cos(rx),-sin(rx));
        rotx[2] = vec3(0.0,-sin(rx),cos(rx));
        mat3 rotz;
        rotz[0] = vec3(cos(rz),0.0,-sin(rz));
        rotz[1] = vec3(0.0,1.0,0.0);
        rotz[2] = vec3(sin(rz),0.0,cos(rz));

        rotatedStem = rotz*rotx*stem;
        stem = rotatedStem;
        root =(root+rotatedStem);
        vPosition0 = (float(i)<calcSegment.x)?(root+rotatedStem):vPosition0;
        vPosition1 = (float(i)<calcSegment.x)?(root+rotatedStem):vPosition1;

      }


      vec3 normal=cross(uWindDir.xyz,normalize(cross(rotatedStem,normRoot))); // not used yet


      vColor = texture2D(uTexture0, vec2(1.0/float(uSize.x),1.0/(float(uSize.y)*2.0))*(vec2(0.5,0.5)+vec2(calcSegment.y,2.0*calcSegment.z+((calcSegment.x<2.5)?0.0:1.0))));

      v_TexCoordinate = texCoordinate;
      vWindForce = uWindDir.xyz*uWindDir.w*length(normal);
      gl_Position = vec4(position,0.0,1.0);
  }`;
        const calculatefragment = `precision highp float;

  uniform float uPhase;
  uniform float uBlockSize;

  uniform vec3 uMidPoints[32];
  uniform float uRadii[32];
  uniform int objectCount;

  varying vec2 v_TexCoordinate;
  varying vec3 vPosition0;
  varying vec3 vPosition1;
  varying vec4 vColor;
  varying vec3 vWindForce;
  varying vec3 vRoot;

  vec3 mod289(vec3 x) {
    return x - floor(x * (1.0 / 289.0)) * 289.0;
  }

  vec4 mod289(vec4 x) {
    return x - floor(x * (1.0 / 289.0)) * 289.0;
  }

  vec4 permute(vec4 x) {
       return mod289(((x*34.0)+1.0)*x);
  }

  vec4 taylorInvSqrt(vec4 r)
  {
    return 1.79284291400159 - 0.85373472095314 * r;
  }

  float snoise(vec3 v)
  {
  const vec2  C = vec2(1.0/6.0, 1.0/3.0) ;
  const vec4  D = vec4(0.0, 0.5, 1.0, 2.0);

// First corner
  vec3 i  = floor(v + dot(v, C.yyy) );
  vec3 x0 =   v - i + dot(i, C.xxx) ;

// Other corners
  vec3 g = step(x0.yzx, x0.xyz);
  vec3 l = 1.0 - g;
  vec3 i1 = min( g.xyz, l.zxy );
  vec3 i2 = max( g.xyz, l.zxy );

  //   x0 = x0 - 0.0 + 0.0 * C.xxx;
  //   x1 = x0 - i1  + 1.0 * C.xxx;
  //   x2 = x0 - i2  + 2.0 * C.xxx;
  //   x3 = x0 - 1.0 + 3.0 * C.xxx;
  vec3 x1 = x0 - i1 + C.xxx;
  vec3 x2 = x0 - i2 + C.yyy; // 2.0*C.x = 1/3 = C.y
  vec3 x3 = x0 - D.yyy;      // -1.0+3.0*C.x = -0.5 = -D.y

// Permutations
  i = mod289(i);
  vec4 p = permute( permute( permute(
             i.z + vec4(0.0, i1.z, i2.z, 1.0 ))
           + i.y + vec4(0.0, i1.y, i2.y, 1.0 ))
           + i.x + vec4(0.0, i1.x, i2.x, 1.0 ));

// Gradients: 7x7 points over a square, mapped onto an octahedron.
// The ring size 17*17 = 289 is close to a multiple of 49 (49*6 = 294)
  float n_ = 0.142857142857; // 1.0/7.0
  vec3  ns = n_ * D.wyz - D.xzx;

  vec4 j = p - 49.0 * floor(p * ns.z * ns.z);  //  mod(p,7*7)

  vec4 x_ = floor(j * ns.z);
  vec4 y_ = floor(j - 7.0 * x_ );    // mod(j,N)

  vec4 x = x_ *ns.x + ns.yyyy;
  vec4 y = y_ *ns.x + ns.yyyy;
  vec4 h = 1.0 - abs(x) - abs(y);

  vec4 b0 = vec4( x.xy, y.xy );
  vec4 b1 = vec4( x.zw, y.zw );

  //vec4 s0 = vec4(lessThan(b0,0.0))*2.0 - 1.0;
  //vec4 s1 = vec4(lessThan(b1,0.0))*2.0 - 1.0;
  vec4 s0 = floor(b0)*2.0 + 1.0;
  vec4 s1 = floor(b1)*2.0 + 1.0;
  vec4 sh = -step(h, vec4(0.0));

  vec4 a0 = b0.xzyw + s0.xzyw*sh.xxyy ;
  vec4 a1 = b1.xzyw + s1.xzyw*sh.zzww ;

  vec3 p0 = vec3(a0.xy,h.x);
  vec3 p1 = vec3(a0.zw,h.y);
  vec3 p2 = vec3(a1.xy,h.z);
  vec3 p3 = vec3(a1.zw,h.w);

//Normalise gradients
  vec4 norm = taylorInvSqrt(vec4(dot(p0,p0), dot(p1,p1), dot(p2, p2), dot(p3,p3)));
  p0 *= norm.x;
  p1 *= norm.y;
  p2 *= norm.z;
  p3 *= norm.w;

// Mix final noise value
  vec4 m = max(0.6 - vec4(dot(x0,x0), dot(x1,x1), dot(x2,x2), dot(x3,x3)), 0.0);
  m = m * m;
  return 42.0 * dot( m*m, vec4( dot(p0,x0), dot(p1,x1),
                                dot(p2,x2), dot(p3,x3) ) );
  }

  float turbulence(vec3 pos,float size)
  {
    float value = 0.0, initialSize = size;
    float size2 = size;
    for(float s =0.0; s<5.0; s+=1.0)
    {
      value += snoise(pos/size2) * size2;
      size2/=2.0;
    }
    return value / initialSize;
  }

  void main ()
  {
    float magnitude = uBlockSize;
    const float multiplier = 0.5;
    const float wmag = 1.0;
      float xt = wmag*vWindForce.x+((turbulence(vPosition0+vec3(0.0,uPhase,0.0), magnitude))*multiplier);
      float yt = wmag*vWindForce.y+((turbulence(vPosition1+vec3(0.0,uPhase+235.0,0.0), magnitude))*multiplier);

      float zt = wmag*vWindForce.x+((turbulence(vPosition0+vec3(0.0,uPhase+475.0,0.0), magnitude)*0.5)*multiplier);
      float wt = wmag*vWindForce.y+((turbulence(vPosition1+vec3(0.0,uPhase+835.0,0.0), magnitude)*0.5)*multiplier);

      for(int i=0; i< 32; i++){
        vec3 dif =uMidPoints[i]-vRoot;
        float r = length(dif);
        vec2 d2 = normalize(dif.xz);
        d2 = (r<uRadii[i])?d2:mix(vec2(xt,yt),d2,1.0-((r-uRadii[i])/(uRadii[i]*2.0)));
        xt = (r<uRadii[i]*3.0 && abs(d2.x)>abs(xt))?d2.x:xt;
        yt = (r<uRadii[i]*3.0 && abs(d2.y)>abs(yt))?-d2.y:yt;
        zt = (r<uRadii[i]*3.0 && abs(d2.x)>abs(zt))?d2.x:zt;
        wt = (r<uRadii[i]*3.0 && abs(d2.y)>abs(wt))?-d2.y:wt;
      }
      vec4 finalColor = (vColor.xyzw*1.0+vec4((xt+1.0)/2.0,(yt+1.0)/2.0,(zt+1.0)/2.0,(wt+1.0)/2.0)*2.0)/3.0;
      gl_FragColor = finalColor;
  }`;
        this.calculateProgram = ShaderLib.makeProgram(gl, calculatevertex, calculatefragment);
        gl.useProgram(this.calculateProgram);
        this.calculateProgram.aPosition = gl.getAttribLocation(this.calculateProgram, 'position');
        this.calculateProgram.aTexCoordinate = gl.getAttribLocation(this.calculateProgram, 'texCoordinate');
        this.calculateProgram.uTexture = gl.getUniformLocation(this.calculateProgram, 'uTexture0');
        this.calculateProgram.uPhase = gl.getUniformLocation(this.calculateProgram, 'uPhase');
        this.calculateProgram.uWindDir = gl.getUniformLocation(this.calculateProgram, 'uWindDir');
        this.calculateProgram.uSize = gl.getUniformLocation(this.calculateProgram, 'uSize');
        this.calculateProgram.aVertexPosition = gl.getAttribLocation(this.calculateProgram, 'inputPosition');
        this.calculateProgram.aCalcSegmentPosition = gl.getAttribLocation(this.calculateProgram, 'calcSegment');
      // calculateProgram.uSampler = gl.getUniformLocation(calculateProgram, 'uSampler');
        this.calculateProgram.uFloorHeight = gl.getUniformLocation(this.calculateProgram, 'uFloorHeight');

        this.calculateProgram.uBlockSize = gl.getUniformLocation(this.calculateProgram, 'uBlockSize');
        this.calculateProgram.uMidPoints = gl.getUniformLocation(this.calculateProgram, 'uMidPoints');
        this.calculateProgram.uRadii = gl.getUniformLocation(this.calculateProgram, 'uRadii');
        this.calculateProgram.uObjectCount = gl.getUniformLocation(this.calculateProgram, 'uObjectCount');
        return this.calculateProgram;
      }


      draw(gl, canvas, perspectiveMatrix, cameraMatrix, view, model) {
        const oldfb = gl.getParameter(gl.FRAMEBUFFER_BINDING);
        gl.useProgram(this.grassprogram);

        gl.uniformMatrix4fv(this.grassprogram.uPerspectiveMatrix, false,
        new Float32Array(perspectiveMatrix.m));
        gl.uniformMatrix4fv(this.grassprogram.uCamera, false, new Float32Array(cameraMatrix.m));
        gl.uniformMatrix4fv(this.grassprogram.uModel, false, new Float32Array(model.m));
        gl.uniformMatrix4fv(this.grassprogram.uView, false, new Float32Array(view.m));
        gl.lineWidth(5);
        gl.activeTexture(this.calctextureunit.unit);
        gl.bindTexture(gl.TEXTURE_2D, this.calctexture);
        gl.uniform1i(gl.getUniformLocation(this.grassprogram, 'uSampler'), this.calctextureunit.number); // todo fix this

        gl.uniform2i(this.grassprogram.uSize, this.w, this.h);
        gl.uniform1f(this.grassprogram.uFloorHeight, this.floorheight);

        gl.bindBuffer(gl.ARRAY_BUFFER, this.gvbuffer);
        gl.enableVertexAttribArray(this.grassprogram.aVertexPosition);
        gl.vertexAttribPointer(this.grassprogram.aVertexPosition, 4, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ARRAY_BUFFER, this.calcsegmentbuffer);
        gl.enableVertexAttribArray(this.grassprogram.aCalcSegmentPosition);
        gl.vertexAttribPointer(this.grassprogram.aCalcSegmentPosition, 4, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.gibuffer);
        for (let i = this.patchesxstart; i < this.patchesxend; i++) {
          for (let j = this.patchesystart; j < this.patchesyend; j++) {
            gl.uniform3f(this.grassprogram.uOffset, i * this.wallx * 1.0, 0, j * this.wallz * 1.0);
            gl.drawElements(gl.TRIANGLE_STRIP, this.w * this.h * 2 * (this.segments + 2), gl.UNSIGNED_INT, 0);
          }
        }
        gl.lineWidth(1);
        if (this.buffernumber == 0) {
          this.calcframebuffer = this.calcframebuffer0;
          this.calctexture = this.calctexture1;
          this.calctextureunit = this.calctextureunit1;
          this.buffernumber = 1;
        } else {
          this.calcframebuffer = this.calcframebuffer1;
          this.calctexture = this.calctexture0;
          this.calctextureunit = this.calctextureunit0;
          this.buffernumber = 0;
        }
        gl.useProgram(this.calculateProgram);
        gl.bindFramebuffer(gl.FRAMEBUFFER, this.calcframebuffer);
        gl.viewport(0, 0, this.w, this.h * 2);

        gl.activeTexture(this.calctextureunit.unit);
        gl.bindTexture(gl.TEXTURE_2D, this.calctexture);
        gl.uniform1i(this.calculateProgram.uTexture, this.calctextureunit.number); // todo fix this
        gl.uniform1f(this.calculateProgram.uPhase, this.phase);
        this.phase += 0.2;
        this.direction += (Math.random() - 0.5) * 0.02;
        this.windDir = [Math.sin(this.direction * Math.PI * 2.0), Math.cos(this.direction * Math.PI * 2.0), 0.0, Math.min(this.windDir[3] + 0.02 * (Math.random() - 0.5), 0.5)];
        gl.bindBuffer(gl.ARRAY_BUFFER, this.vbuffer);
        gl.enableVertexAttribArray(this.calculateProgram.aPosition);
        gl.vertexAttribPointer(this.calculateProgram.aPosition, 2, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ARRAY_BUFFER, this.tbuffer);
        gl.enableVertexAttribArray(this.calculateProgram.aTexCoordinate);
        gl.vertexAttribPointer(this.calculateProgram.aTexCoordinate, 2, gl.FLOAT, false, 0, 0);

        gl.uniform2i(this.calculateProgram.uSize, this.w, this.h);
        gl.uniform1f(this.calculateProgram.uFloorHeight, this.floorheight);

        gl.uniform1f(this.calculateProgram.uBlockSize, this.wallx / 10.0);
        gl.uniform4fv(this.calculateProgram.uWindDir, this.windDir);

        gl.uniform3fv(this.calculateProgram.uMidPoints, this.mMidPoints);
        gl.uniform1fv(this.calculateProgram.uRadii, this.mRadii);
        gl.uniform1i(this.calculateProgram.uObjectCount, this.mObjectCount);

        gl.bindBuffer(gl.ARRAY_BUFFER, this.gvbuffer);
        gl.enableVertexAttribArray(this.calculateProgram.aVertexPosition);
        gl.vertexAttribPointer(this.calculateProgram.aVertexPosition, 4, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ARRAY_BUFFER, this.calcsegmentbuffer);
        gl.enableVertexAttribArray(this.calculateProgram.aCalcSegmentPosition);
        gl.vertexAttribPointer(this.calculateProgram.aCalcSegmentPosition, 4, gl.FLOAT, false, 0, 0);
        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.gi2buffer);

        gl.drawElements(gl.POINTS, this.w * this.h * 2, gl.UNSIGNED_INT, 0);
      // gl.drawArrays(gl.POINTS,0,w*h*2);
        gl.bindFramebuffer(gl.FRAMEBUFFER, oldfb);
        gl.viewport(0, 0, canvas.width, canvas.height);
      }

    }

    GrassLib.Grass = Grass;
    return GrassLib;
  }
// define globally if it doesn't already exist
  if (typeof (GrassLib) === 'undefined') {
    const mwindow = window;
    mwindow.GrassLib = defineLibrary();
  }
}(window)); // eslint-disable-line no-undef
