(function capsule(window) {
  function defineLibrary() {
    const PlantsLib = {};
    const Vec3 = VemaLib.Vec3;
    const Matrix4 = VemaLib.Matrix4;


    const treeElementType = {
      cube: 0,
      square: 1,
      triangle: 2,
      salmiak: 3,
      cone: 4,
      none: 5
    };


    const treetriangle = [
      0.5, 0, 0,
      0, 1, 0,
      -0.5, 0, 0
    ];

    const treetrianglen = [
      0, 0, 1,
      0, 0, 1,
      0, 0, 1
    ];

    const treesalmiak = [
      -0.5, 0.5, 0,
      0, 0, 0,
      0, 1, 0,
      0.5, 0.5, 0
    ];

    const treesalmiakn = [
      0, 0, 1,
      0, 0, 1,
      0, 0, 1,
      0, 0, 1
    ];

    const treesquare = [
      -0.5, 1, 0,
      -0.5, 0, 0,
      0.5, 1, 0,
      0.5, 0, 0
    ];

    const treesquaren = [
      0, 0, 1,
      0, 0, 1,
      0, 0, 1,
      0, 0, 1,
    ];

    const treecone = [

      -0.5, 0, -0.5,
      0, 1, 0,

      0.5, 0, -0.5,
      0, 1, 0,

      0.5, 0, 0.5,
      0, 1, 0,

      -0.5, 0, 0.5,
      0, 1, 0,
      -0.5, 0, -0.5,

    ];

    const tca = 1 / Math.sqrt(10);
    const tcb = 2 / Math.sqrt(5);
    const treeconen = [


      -tca, tcb, -tca,
      -tca, tcb, -tca,

      tca, tcb, -tca,
      tca, tcb, -tca,


      tca, tcb, tca,
      tca, tcb, tca,

      -tca, tcb, tca,
      -tca, tcb, tca,
      -tca, tcb, -tca,


    ];

    const treecube = [

      -0.5, 1, -0.5,
      -0.5, 0, -0.5,
      0.5, 1, -0.5,
      0.5, 0, -0.5,

      0.5, 0, -0.5,
      0.5, 1, -0.5,
      0.5, 0, 0.5,
      0.5, 1, 0.5,

      0.5, 1, 0.5,
      0.5, 0, 0.5,
      -0.5, 1, 0.5,
      -0.5, 0, 0.5,

      -0.5, 0, 0.5,
      -0.5, 1, 0.5,
      -0.5, 0, -0.5,
      -0.5, 1, -0.5,

      -0.5, 1, -0.5,
      -0.5, 1, 0.5,
      0.5, 1, -0.5,
      0.5, 1, 0.5,

    ];

    const treecuben = [

      0, 0, -1,
      0, 0, -1,
      0, 0, -1,
      0, 0, -1,

      1, 0, 0,
      1, 0, 0,
      1, 0, 0,
      1, 0, 0,

      0, 0, 1,
      0, 0, 1,
      0, 0, 1,
      0, 0, 1,

      -1, 0, 0,
      -1, 0, 0,
      -1, 0, 0,
      -1, 0, 0,

      0, 1, 0,
      0, 1, 0,
      0, 1, 0,
      0, 1, 0,
      0, 1, 0,
    ];

    const elemProps = {
      0: { vertices: treecube, numVertices: 20, normals: treecuben },
      1: { vertices: treesquare, numVertices: 4, normals: treesquaren },
      2: { vertices: treetriangle, numVertices: 3, normals: treetrianglen },
      3: { vertices: treesalmiak, numVertices: 4, normals: treesalmiakn },
      4: { vertices: treecone, numvertices: 9, normals: treeconen },
      5: { vertices: 0, numVertices: 0, normals: 0 }

    };

    PlantsLib.generator = new Random.Generator(12433, 46831, 66713, 11447);

    class ProtoBranch {
      constructor(maxdepth, depth, position, up, normal, height, width, ax, ay, az, childBranch) {
        this.maxdepth = maxdepth;
        this.depth = depth;
        this.position = position;
        this.width = width;
        this.height = height;
        this.anglex = ax;
        this.anglez = az;
        this.angley = ay;
        this.normal = normal;
        this.up = up;
        this.childBranch = childBranch;
        this.elementType = treeElementType.cube;
        const shadered = (maxdepth - (maxdepth - depth)) / (maxdepth * 6) + 0.1;
        const shadegreen = (maxdepth - depth) / (maxdepth * 4) + 0.2;
        const shadeblue = 0.2;
        this.color = [shadered, shadegreen, shadeblue];
      }

      stem(subbranches, xadd, yadd, zadd, len0, len1, wscale) {
        const right = this.normal.cross(this.up).normalized();
        const rotz = new Matrix4();
        rotz.rotateMatrix(zadd, this.normal);
        const roty = new Matrix4();
        roty.rotateMatrix(yadd, this.up);
        const rotx = new Matrix4();
        rotx.rotateMatrix(xadd, right);


        const rot = rotx.multiple(roty.multiple(rotz));
        const len = this.height * len1;
        const dir = rot.multipleV(this.up.multiple(this.height * len1));
        const nup = dir.normalized();
        const pos = this.position.add(this.up.multiple(this.height * len0));
        const nor = rot.multipleV(this.normal).normalized();


        // const pp = position.add(dir2);
        subbranches.push(new this.childBranch(this.maxdepth, this.depth - 1, pos, nup, nor, len, this.width * wscale, xadd + this.anglex, yadd + this.angley, zadd + this.anglez, this.childBranch));
      }
    }

    class FirBranch extends ProtoBranch {
      constructor(maxdepth, depth, position, up, normal, height, width, ax, ay, az, childBranch) {
        super(maxdepth, depth, position, up, normal, Math.max(height, 0.02), Math.max(width, 0.02), ax, ay, az, childBranch);
        this.elementType = treeElementType.square;
        const shadegreen = (maxdepth - depth) / (maxdepth * 5) + 0.1;
        const shadeblue = 0.2;
        this.color = [shadeblue, shadegreen, shadeblue];
        if (depth > 0)    		{
          this.subbranches = [];
          this.mainstem(this.subbranches);
          this.leftstem(this.subbranches);
          this.rightstem(this.subbranches);
        }
      }
      mainstem(subbranches) {
        super.stem(subbranches, (PlantsLib.generator.random() - 0.5) * 0.2, (PlantsLib.generator.random() - 0.5) * 0.2, (PlantsLib.generator.random() - 0.5) * 0.2 + 0.15,
        1.0, 0.8, 0.825);
      }

      leftstem(subbranches) {
        super.stem(subbranches, Math.PI * 3 / 8 + (PlantsLib.generator.random() - 0.5) * 0.2, (PlantsLib.generator.random() - 0.5) * 0.02, 0.5 + (PlantsLib.generator.random() - 0.5) * 0.2,
        0.9 + PlantsLib.generator.random() * 0.1, 0.5 + PlantsLib.generator.random() * 0.1, 0.925);
      }

      rightstem(subbranches) {
        super.stem(subbranches, -Math.PI * 3 / 8 + (PlantsLib.generator.random() - 0.5) * 0.2, (PlantsLib.generator.random() - 0.5) * 0.02, 0.5 + (PlantsLib.generator.random() - 0.5) * 0.2,
        0.9 + PlantsLib.generator.random() * 0.1, 0.5 + PlantsLib.generator.random() * 0.1, 0.925);
      }

    }

    class FirTrunk extends ProtoBranch {
      constructor(maxdepth, depth, position, up, normal, height, width, ax, ay, az, childBranch) {
        super(maxdepth, depth, position, up, normal, height, width, ax, ay, az, childBranch);
        this.ff = 0.3;
        // const shadered = (maxdepth-(maxdepth-depth))/(maxdepth*6)+0.1;
        const shadegreen = (maxdepth - depth) / (maxdepth * 5) + 0.1;
        const shadeblue = 0.2;
        this.color = [shadeblue, shadegreen, shadeblue];
        if (depth > 0) {
          this.subbranches = [];
          this.mainstem(this.subbranches);
          this.childBranch = FirBranch;
          const r = 3 + Math.floor(PlantsLib.generator.random() * 3.5);
          for (let i = 0; i < r; i++) {
            const turn = new Matrix4();
            turn.rotateMatrix(Math.PI * 2 / r + (PlantsLib.generator.random() - 0.5) * 0.5 * Math.PI, up);

            this.normal = turn.multipleV(this.normal);
            this.stem(this.subbranches);
            this.halfstem(this.subbranches);
          }
        }
        console.log(position.f);
      }
      mainstem(subbranches) {
        super.stem(subbranches, (PlantsLib.generator.random() - 0.5) * 0.02, (PlantsLib.generator.random() - 0.5) * 0.02, (PlantsLib.generator.random() - 0.5) * 0.02,
          1.0, 0.78, 0.78);
      }

      stem(subbranches) {
        super.stem(subbranches, (PlantsLib.generator.random() - 0.5) * this.ff, (PlantsLib.generator.random() - 0.5) * this.ff, Math.PI * 3 / 8 + (PlantsLib.generator.random() - 0.5) * this.ff,
          0.6 + PlantsLib.generator.random() * 0.4, 0.4, 0.4);
      }

      halfstem(subbranches) {
        super.stem(subbranches, (PlantsLib.generator.random() - 0.5) * this.ff, (PlantsLib.generator.random() - 0.5) * this.ff, Math.PI * 3 / 8 + (PlantsLib.generator.random() - 0.5) * this.ff,
          0.2 + PlantsLib.generator.random() * 0.15, 0.4, 0.4);
      }


    }


    class Fir {
      constructor(depth, position, height) {
        this.pos = position;
        this.trunk = new FirTrunk(depth, depth, position.add(new Vec3(0, 0, 0)), new Vec3(0, 1, 0), new Vec3(0, 0, 1), height, height / 8, 0, 0, 0, FirTrunk);
      }
    }


    class Spruce {
      constructor(gl, canvas, position) {
        this.t = 0;
        this.vb = `
    attribute vec3 inputPosition;
    attribute vec3 inputNormal;
    attribute vec4 uColor;
    	uniform mat4 projection, modelview, uL, uC;
    varying vec4 color;
    //varying vec3 u;
    varying vec3 n;
    	void main(){

        color=uColor;
        //u=(uC*modelview*vec4(inputPosition,1.)).xyz;
        n=normalize((modelview*vec4(inputNormal,0.)).xyz);
        gl_Position = projection*uC*uL * modelview * vec4(inputPosition,1.);
    	}
    `;

        this.fb =
    `precision mediump float;
    	varying vec4 color;
      //varying vec3 u;
      varying vec3 n;
      uniform vec3 k;
      const float screenGamma =2.2;

      void main() {
         vec4 texCol = vec4(1.0,1.0,1.0,1.0);
         vec4 c = color;

          vec3 ambientColor = c.xyz/4.0;
        vec3 diffuseColor = c.xyz;
          vec3 normal = normalize(n);
        vec3 lightDir = normalize(k);

          float lambertian = max(abs(dot(normal,lightDir)), 0.0);

          vec4 colorLinear = vec4(ambientColor +
                             lambertian * diffuseColor,1.);
          gl_FragColor = colorLinear;
          //vec4 colorGammaCorrected = vec4(pow(colorLinear, vec3(1.0/screenGamma)),1.0);
          //float vd=clamp((-20.+length(k-u))/20.,0.,1.);
          //gl_FragColor=mix(colorLinear,vec4(.9,.9,.9,1.),vd);
        }`;

        this.program = ShaderLib.makeProgram(gl, this.vb, this.fb);
        gl.useProgram(this.program);
        this.params = {};
        this.params.projection = gl.getUniformLocation(this.program, 'projection');
        this.params.modelview = gl.getUniformLocation(this.program, 'modelview');
        this.params.uC = gl.getUniformLocation(this.program, 'uC');
        this.params.uL = gl.getUniformLocation(this.program, 'uL');
        this.params.k = gl.getUniformLocation(this.program, 'k');
        this.params.uColor = gl.getAttribLocation(this.program, 'uColor');
        this.params.aVertexPosition = gl.getAttribLocation(this.program, 'inputPosition');
        this.params.aNormalPosition = gl.getAttribLocation(this.program, 'inputNormal');
        this.tree = new Fir(11, new Vec3(position[0], position[1], position[2]), 1.0);
        this.vertexBuffer = [];
        this.colorBuffer = [];
        this.normalBuffer = [];
        this.position = position;
        this.prerenderTree(gl, canvas, this.tree, this.vertexBuffer, this.colorBuffer, this.normalBuffer);
        this.renderTree(gl, canvas);
        this.len = this.vertexBuffer.length / 3;
        console.log(`vertices ${this.len}`);
      }

      recursivePrerenderTree(bra, treeRowData, treeRowColorData, treeRowNormalData) {
        if (bra) {
          const right = bra.normal.cross(bra.up).normalized();
          const rotz = new Matrix4();
          rotz.rotateMatrix(bra.anglez, bra.normal);

          const roty = new Matrix4();
          roty.rotateMatrix(bra.angley, bra.up);

          const rotx = new Matrix4();
          rotx.rotateMatrix(bra.anglex, right);

          const rot = rotx.multiple(roty.multiple(rotz));

          const scale = new Matrix4();
          scale.scale(bra.width, bra.height, bra.width);
          const roto = scale.multiple(rot);
          const elem = elemProps[bra.elementType];
          for (let i = 0; i < elem.numVertices; i++) {
            const dir2 = new Vec3(elem.vertices[i * 3], elem.vertices[i * 3 + 1], elem.vertices[i * 3 + 2]);
            const dir = roto.multipleV(dir2);
            const nor = rot.multipleV(new Vec3(elem.normals[i * 3], elem.normals[i * 3 + 1], elem.normals[i * 3 + 2]));
            const posi = dir.add(bra.position);
            if (i === 0 || i === elem.numVertices - 1) {
              treeRowData.push(posi.f[0]);
              treeRowData.push(posi.f[1]);
              treeRowData.push(posi.f[2]);
              treeRowNormalData.push(nor.f[0]);
              treeRowNormalData.push(nor.f[1]);
              treeRowNormalData.push(nor.f[2]);
              treeRowColorData.push(bra.color[0]);
              treeRowColorData.push(bra.color[1]);
              treeRowColorData.push(bra.color[2]);
              treeRowColorData.push(1.0);
            }
            treeRowData.push(posi.f[0]);
            treeRowData.push(posi.f[1]);
            treeRowData.push(posi.f[2]);
            treeRowNormalData.push(nor.f[0]);
            treeRowNormalData.push(nor.f[1]);
            treeRowNormalData.push(nor.f[2]);
            treeRowColorData.push(bra.color[0]);
            treeRowColorData.push(bra.color[1]);
            treeRowColorData.push(bra.color[2]);
            treeRowColorData.push(1.0);
          }
          if (bra.subbranches) {
            for (let i = 0; i < bra.subbranches.length; i++) {
              this.recursivePrerenderTree(bra.subbranches[i], treeRowData, treeRowColorData, treeRowNormalData);
            }
          }
        }
      }


      prerenderTree(gl, canvas, tree, treeRowData, treeRowColorData, treeRowNormalData) {
        this.recursivePrerenderTree(tree.trunk, treeRowData, treeRowColorData, treeRowNormalData);
      }

      renderTree(gl, canvas) {
        gl.useProgram(this.program);
        this.vertices = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, this.vertices);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(this.vertexBuffer), gl.STATIC_DRAW);
        this.colors = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, this.colors);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(this.colorBuffer), gl.STATIC_DRAW);
        this.normals = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, this.normals);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(this.normalBuffer), gl.STATIC_DRAW);
      }

      drawTree(gl, canvas) {
        gl.bindBuffer(gl.ARRAY_BUFFER, this.vertices);
        gl.enableVertexAttribArray(this.params.aVertexPosition);
        gl.vertexAttribPointer(this.params.aVertexPosition, 3, gl.FLOAT, false, 0, 0);
        gl.bindBuffer(gl.ARRAY_BUFFER, this.normals);
        gl.enableVertexAttribArray(this.params.aNormalPosition);
        gl.vertexAttribPointer(this.params.aNormalPosition, 3, gl.FLOAT, false, 0, 0);
        gl.bindBuffer(gl.ARRAY_BUFFER, this.colors);
        gl.enableVertexAttribArray(this.params.uColor);
        gl.vertexAttribPointer(this.params.uColor, 4, gl.FLOAT, false, 0, 0);
        gl.drawArrays(gl.TRIANGLE_STRIP, 0, this.len);
      }


      draw(gl, canvas, pm, vm, cm, mm) {
        this.light = new Float32Array([5.0, 40.0, -3]);
        gl.useProgram(this.program);

        gl.uniformMatrix4fv(this.params.projection, false, new Float32Array(pm.m));
        gl.uniformMatrix4fv(this.params.modelview, false, new Float32Array(mm.m));
        gl.uniformMatrix4fv(this.params.uC, false, new Float32Array(cm.m));
        gl.uniformMatrix4fv(this.params.uL, false, new Float32Array(vm.m));
        gl.uniform3fv(this.params.k, this.light);
        this.drawTree(gl, canvas);
      }

      setT(t) {
        this.t = t;
      }

    }

    PlantsLib.Spruce = Spruce;

    return PlantsLib;
  }
// define globally if it doesn't already exist
  if (typeof (PlantsLib) === 'undefined') {
    const mwindow = window;
    mwindow.PlantsLib = defineLibrary();
  }
}(window)); // eslint-disable-line no-undef
