(function capsule(window) {
  function defineLibrary() {
    const Puzzle = {};

    Puzzle.PuzzleFilter = class PuzzleFilter
      extends FilterLib.Filter {
        getVertexShader() {
          this.vertexShader = `
precision highp float;
uniform vec2 middlepoint;
uniform vec2 randompoint;
uniform vec2 uSize;
uniform float stage;
uniform float randomrotation;
attribute vec2 position;
attribute vec2 texCoordinate;
varying vec2 v_TexCoordinate;

void main(){
  v_TexCoordinate = texCoordinate;
  vec2 ratio = vec2(1.0,uSize.y/uSize.x);
  vec4 mp = vec4(middlepoint*ratio,0.0,1.0);
  vec4 rp = vec4(randompoint*ratio,0.0,1.0);
  vec4 pos =vec4(position*ratio,0.0,1.0);
  float rrot = mix(0.0,randomrotation,stage);
  mat4 mrot = mat4(cos(rrot),-sin(rrot),0.0,0.0,
                  sin(rrot),cos(rrot),0.0,0.0,
                  0.0,0.0,1.0,0.0,
                  0.0,0.0,0.0,1.0);
  vec4 tmp = mix(mp,rp,stage);
  vec4 tdiff=pos-mp;
  vec4 tpos =tdiff*mrot+tmp;
  gl_Position = vec4(tpos.xy/ratio,0.0,1.0);
}`;
        }

        getFragmentShader() {
          this.fragmentShader = `
precision highp float;
uniform int black;
uniform ivec2 cxy;
uniform sampler2D uTexture0;
uniform float uSetScalar0;
uniform float uSetScalar1;
uniform vec2 uSize;
varying vec2 v_TexCoordinate;

// Description : Array- and textureless GLSL 2D simplex noise.
// Author : Ian McEwan, Ashima Arts. Version: 20110822
// Copyright (C) 2011 Ashima Arts. All rights reserved.
// Distributed under the MIT License. See LICENSE file.
// https://github.com/ashima/webgl-noise

vec2 mod289(vec2 x) { return x - floor(x * (1.0 / 289.0)) * 289.0; }
vec3 mod289(vec3 x) { return x - floor(x * (1.0 / 289.0)) * 289.0; }
vec3 permute(vec3 x) { return mod289((( x * 34.0) + 1.0) * x); }

float snoise(vec2 v) {
const vec4 C = vec4(0.211324865405187,  // (3.0-sqrt(3.0))/6.0
                  0.366025403784439,  // 0.5*(sqrt(3.0)-1.0)
                 -0.577350269189626,  // -1.0 + 2.0 * C.x
                  0.024390243902439); // 1.0 / 41.0
// First corner
vec2 i = floor(v + dot(v, C.yy) );
vec2 x0 = v - i + dot(i, C.xx);
// Other corners
vec2 i1 = (x0.x > x0.y) ? vec2(1.0, 0.0) : vec2(0.0, 1.0);
vec4 x12 = x0.xyxy + C.xxzz;
x12.xy -= i1;
// Permutations
i = mod289(i); // Avoid truncation effects in permutation
vec3 p = permute( permute( i.y + vec3(0.0, i1.y, 1.0 ))
                       + i.x + vec3(0.0, i1.x, 1.0 ));
vec3 m = max(0.5 - vec3(dot(x0,x0), dot(x12.xy,x12.xy),
                      dot(x12.zw,x12.zw)), 0.0);
m = m*m; m = m*m;
// Gradients
vec3 x = 2.0 * fract(p * C.www) - 1.0;
vec3 h = abs(x) - 0.5;
vec3 a0 = x - floor(x + 0.5);
// Normalise gradients implicitly by scaling m
m *= 1.792843 - 0.853735 * ( a0*a0 + h*h );
// Compute final noise value at P
vec3 g;
g.x = a0.x * x0.x + h.x * x0.y;
g.yz = a0.yz * x12.xz + h.yz * x12.yw;
return 130.0 * dot(m, g);
}

float random (vec2 st) {
return fract(sin(dot(st.xy,
                 vec2(12.9898,78.233)))*
43758.5453123);
}

vec4 Puzzle(vec2 uvw, float lw) {
vec4 finalcolor= vec4(0.0,0.0,0.0,0.0);
for(int i=-1;i<2;i++)
for(int j=-1;j<2;j++){
vec2 uvws =uvw+lw*vec2(float(i),float(j))/3.0;
vec2 uvwr =uvws+0.075*snoise(0.6*uvw);
vec2 uvwi = floor(uvwr); // cell coords
vec2 uvwd= uvwr-uvwi;

float me=random(uvwi);
float upcell = random(floor(uvwi+vec2(0.0,1.0))); // cell ID
float rightcell = random(floor(uvwi+vec2(1.0,0.0))); // cell ID
int both =(me<0.25)?1:0;
bool bottom = (me<0.5);
bool left = ((me>=0.5&&me<0.75)||me<0.25);
bool right = !((rightcell>=0.5&&rightcell<0.75)||rightcell<0.25);
bool top = !(upcell<0.5);

const float pw = 0.15;  //pegwidth
const float cr = 0.16; // circle radius;
const float pw2 = pw/2.0; //halfpegwidth
const float ph = 0.05; //peg height
const float ph2 = ph/2.0;
const float ptc = (cr-cos(asin(pw2/cr))*cr)*0.5;//peg to circle halved

//
vec2 ycmp = uvwd-vec2(0.5,cr+ph);
float ycircled = length(ycmp);
bool yinsideCircle = bottom&&(ycircled<cr);
bool yinCircle = yinsideCircle&&ycircled>cr-lw;

vec2 ypegw = uvwd-vec2(0.5,ph2+ptc);
bool yPegged = (abs(ypegw.x)<pw2);
bool yPeggedHigh= bottom&&(abs(ypegw.y)<(ph2+ptc+lw));
bool yPeggedInside =(abs(ypegw.x)>(pw2-lw)&& yPeggedHigh);

//
vec2 yicmp = uvwd-vec2(0.5,(1.0-cr-ph));
float yicircled = length(yicmp);
bool yiinsideCircle = top&&(yicircled<cr);
bool yiinCircle = yiinsideCircle&&yicircled>cr-lw;

vec2 yipegw = uvwd-vec2(0.5,1.0-(ph2+ptc));
bool yiPegged = (abs(yipegw.x)<pw2);
bool yiPeggedHigh= top&&(abs(yipegw.y)<(ph2+ptc+lw));
bool yiPeggedInside =(abs(yipegw.x)>((pw2-lw))&& yiPeggedHigh);

//
float xcircled = length((uvwd-vec2(cr+ph,0.5)).xy);
bool xinsideCircle = left&&xcircled<cr;
bool xinCircle = xinsideCircle&&xcircled>cr-lw;

vec2 xpegw = uvwd-vec2(ph2+ptc,0.5);
bool xPegged = (abs(xpegw.y)<pw2);
bool xPeggedHigh= left&&(abs(xpegw.x)<(ph2+ptc+lw));
bool xPeggedInside =(abs(xpegw.y)>(pw2-lw)&& xPeggedHigh);

//
float xicircled = length((uvwd-vec2((1.0-(cr+ph)),0.5)));
bool xiinsideCircle = right&&xicircled<cr;
bool xiinCircle = xiinsideCircle&&xicircled>cr-lw;

vec2 xipegw = uvwd-vec2(1.0-(ph2+ptc),0.5);
bool xiPegged = (abs(xipegw.y)<pw2);
bool xiPeggedHigh= right&&(abs(xipegw.x)<(ph2+ptc+lw));
bool xiPeggedInside =(abs(xipegw.y)>(pw2-lw)&& xiPeggedHigh);


bool borderx =(uvwd.x>lw||(uvwd.y<0.5+pw2-lw&&uvwd.y > 0.5-pw2+lw));
bool bordery = (uvwd.y>lw||(uvwd.x<0.5+pw2-lw&&uvwd.x > 0.5-pw2+lw));


vec4 currentcolor=(((borderx &&
bordery && !yinCircle &&
!xinCircle && !yiinCircle && !xiinCircle)|| ( yPeggedHigh && yPegged)||
( xPeggedHigh&& xPegged)|| ( yiPeggedHigh&& yiPegged)||
( xiPeggedHigh&& xiPegged)) && (!yPegged || !yPeggedInside)&&
(!xPegged ||!xPeggedInside)&& (!yiPegged ||!yiPeggedInside)&&
(!xiPegged ||!xiPeggedInside))?texture2D(uTexture0,v_TexCoordinate):vec4(0.0,0.0,0.0,1.0);


if(int(uvwi.x)==cxy.x && int(uvwi.y) == cxy.y && !(currentcolor.r <0.01)&&
(
  (left && (xinsideCircle || (xPegged && xPeggedHigh))) ||
  (right && (xiinsideCircle|| (xiPegged && xiPeggedHigh))) ||
  (bottom && (yinsideCircle|| (yPegged && yPeggedHigh))) ||
  (top && (yiinsideCircle|| (yiPegged && yiPeggedHigh)))
)
){
   discard;
 }
if(

  (!(left && (xinsideCircle|| (xPegged && xPeggedHigh)))
  && int(uvwi.x)==cxy.x+1 && int(uvwi.y) == cxy.y)&&borderx
  ||
  (!(right&&(xiinsideCircle|| (xiPegged && xiPeggedHigh)))
  && int(uvwi.x)==cxy.x-1 && int(uvwi.y) == cxy.y)

  ||

  (
    (!(bottom && (yinsideCircle|| (yPegged && yPeggedHigh)))
    && int(uvwi.x)==cxy.x && int(uvwi.y) == cxy.y+1)&&bordery
  )

  ||
  (
    (!(top && (yiinsideCircle|| (yiPegged && yiPeggedHigh)))
    && int(uvwi.x)==cxy.x && int(uvwi.y) == cxy.y-1)

  )

  ){
  discard;
}

if(((int(uvwi.y) != cxy.y ||int(uvwi.x)+1<cxy.x || int(uvwi.x)-1>cxy.x))&& (int(uvwi.x) != cxy.x ||int(uvwi.y) +1< cxy.y || int(uvwi.y) -1> cxy.y)){
  currentcolor = mix(vec4(0.8,0.5,0.5,1.0),currentcolor,0.5);
  discard;
}
finalcolor+=currentcolor;
}
return finalcolor/9.0;
}

void main () {
vec4 cameraColor = vec4(1.0,1.0,1.0,1.0);
vec2 ratio = vec2(1.0,uSize.y/uSize.x);
float scale = (uSetScalar0*20.0+1.0);
if(black ==1){
   gl_FragColor = vec4(0.1,0.3,0.1,1.0);
   return;
}
vec4 finalColor = Puzzle((v_TexCoordinate-vec2(0.5,0.5))*ratio*scale,uSetScalar1*2.0*scale/uSize.x);
gl_FragColor = vec4(finalColor.rgb,1.0);
}`;
        }

        constructor(gl, canvas) {
          super(gl, canvas);
          this.mIsOnOffFilter = true;
          this.mIsSizeFilter = true;
          this.mIsSetUniformsFilter = true;
          this.mIsSpecialRenderer = true;
          this.mIsStartFilter = true;
          this.mIsCalculateFilter = true;
          this.currentscalars = [0.5, 0.5, 0.5, 0.5];
          this.mScalars = [0.5, 0.5, 0.5, 0.5];
          this.cx = 2;
          this.cy = 2;
          this.pwidth = 10;
          this.pheight = 10;
          this.startFilter(gl);
          this.updateScalars();
        }

        resize(gl, canvas) {
          this.setSize(canvas.width, canvas.height);
          super.resize(gl, canvas);
        }

        setSize(width, height) {
          if (this.mWidth !== width || this.mHeight !== height) {
            this.mWidth = width;
            this.mHeight = height;
            this.updateScalars();
          }
        }

        setScalar(i, scalar) {
          if (this.currentscalars[i] !== scalar && i < 2) {
            this.currentscalars[i] = this.mScalars[i] = scalar;
            this.updateScalars();
          } else {
            this.currentscalars[i] = this.mScalars[i] = scalar;
          }
        }


        updateScalars() {
          this.ratio = [1.0, this.mWidth / this.mHeight];
          this.scale = (0.5 * 20.0 + 1.0); // this.scale = (this.currentscalars[0] * 20.0 + 1.0);
          this.pwidth = Math.ceil(this.scale) + 1;
          this.pheight = Math.ceil(this.scale) + 1;
          // console.log(`width ${this.pwidth} height ${this.pheight}`);
          this.cx = Math.floor((1.0) / this.ratio[0] * this.scale);
          this.cy = Math.ceil((1.0) / this.ratio[1] * this.scale);
          this.needsupdate = true;
        }

        setExtraParams(gl) {
          gl.uniform1i(this.params.black, 1);
          gl.uniform1f(this.params.randomrotation, 0.0);
          gl.uniform2f(this.params.middlepoint, 0.0, 0.0);
          gl.uniform2f(this.params.randompoint, 0.0, 0.0);
          gl.uniform2i(this.params.cxy, -100, -100);
          gl.uniform1f(this.params.stage, 0.0);
        }

        draw(gl, canvas, fb) {
          this.bindFrameBuffer(gl);
          gl.useProgram(this.program);
          gl.viewport(0, 0, canvas.width, canvas.height);
          gl.clearColor(0.2, 0.5, 0.1, 1.0);
          // Enable depth testing
          gl.enable(gl.DEPTH_TEST);
          // Near things obscure far things
          gl.depthFunc(gl.LEQUAL);
          // Clear the color as well as the depth buffer.
          gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT); // eslint-disable-line no-bitwise

          if (this.needsupdate) {
            const coords = new Float32Array(2 * 4 * (this.pwidth) * (this.pheight));
            const texcoords = new Float32Array(2 * 4 * (this.pwidth) * (this.pheight));
            const indices = new Uint16Array(6 * this.pwidth * this.pheight);

            let offset = 0;
            let toffset = 0;
            let ioffset = 0;
            this.halfpwidth = Math.floor(this.pwidth / 2);
            this.halfpheight = Math.floor(this.pheight / 2);
            // this.middlepoints = [];
            // this.randompoints = [];
            // this.randomrotations = [];
            // this.atstartpoint = [];
            this.pieces = [];
            for (let j = 0; j < this.pheight; j++) {
              for (let i = 0; i < this.pwidth; i++) {
                let sumx = 0.0;
                let sumy = 0.0;
                sumx += coords[offset++] = (-this.pwidth + 2.0 * (i - 1)) / this.scale;
                sumy += coords[offset++] = (-this.pheight + 2.0 * (j - 1)) / (this.scale / this.ratio[1]);
                sumx += coords[offset++] = (-this.pwidth + 2.0 * (i + 2)) / this.scale;
                sumy += coords[offset++] = (-this.pheight + 2.0 * (j - 1)) / (this.scale / this.ratio[1]);
                sumx += coords[offset++] = (-this.pwidth + 2.0 * (i - 1)) / this.scale;
                sumy += coords[offset++] = (-this.pheight + 2.0 * (j + 2)) / (this.scale / this.ratio[1]);
                sumx += coords[offset++] = (-this.pwidth + 2.0 * (i + 2)) / this.scale;
                sumy += coords[offset++] = (-this.pheight + 2.0 * (j + 2)) / (this.scale / this.ratio[1]);
                // this.middlepoints.push([sumx / 4.0, sumy / 4.0]);
                // this.randompoints.push([(Random.random() - 0.5) * 2.0, (Random.random() - 0.5) * 2.0]);
                // this.randomrotations.push(Math.PI * 12.0 * Random.random() - Math.PI * 6.0);
                // this.atstartpoint.push(Random.random() * 0.99);
                this.pieces.push({ index: i + j * this.pwidth,
                  middlepoint: [sumx / 4.0, sumy / 4.0],
                  randompoint: [(Random.random() - 0.5) * 2.0, (Random.random() - 0.5) * 2.0],
                  randomrotation: Math.PI * 12.0 * Random.random() - Math.PI * 6.0,
                  atstartpoint: Random.random() * 0.8,
                  i,
                  j });
                texcoords[toffset++] = 0.5 + (-this.halfpwidth + (i - 1)) / this.scale;
                texcoords[toffset++] = 0.5 + (-this.halfpheight + (j - 1)) / (this.scale / this.ratio[1]);
                texcoords[toffset++] = 0.5 + (-this.halfpwidth + (i + 2)) / this.scale;
                texcoords[toffset++] = 0.5 + (-this.halfpheight + (j - 1)) / (this.scale / this.ratio[1]);
                texcoords[toffset++] = 0.5 + (-this.halfpwidth + (i - 1)) / this.scale;
                texcoords[toffset++] = 0.5 + (-this.halfpheight + (j + 2)) / (this.scale / this.ratio[1]);
                texcoords[toffset++] = 0.5 + (-this.halfpwidth + (i + 2)) / this.scale;
                texcoords[toffset++] = 0.5 + (-this.halfpheight + (j + 2)) / (this.scale / this.ratio[1]);
              }
            }
            this.pieces.sort((a, b) => a.atstartpoint - b.atstartpoint);


            for (let i = 0; i < this.pieces.length; i++) {
              const koffset = this.pieces[i].index * 4;
              indices[ioffset++] = koffset;
              indices[ioffset++] = koffset + 1;
              indices[ioffset++] = koffset + 2;
              indices[ioffset++] = koffset + 2;
              indices[ioffset++] = koffset + 1;
              indices[ioffset++] = koffset + 3;
            }

            gl.bindBuffer(gl.ARRAY_BUFFER, this.vbuffer);
            gl.bufferData(gl.ARRAY_BUFFER, coords, gl.STATIC_DRAW);
            gl.bindBuffer(gl.ARRAY_BUFFER, this.tbuffer);
            gl.bufferData(gl.ARRAY_BUFFER, texcoords, gl.STATIC_DRAW);
            gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.ibuffer);
            gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, indices, gl.STATIC_DRAW);
            this.needsupdate = false;
          }
          gl.activeTexture(fb.textureUnit.unit);
          gl.bindTexture(gl.TEXTURE_2D, fb.texture);
          gl.uniform1i(this.params.uTexture0, fb.textureUnit.number);

          gl.uniform1i(this.params.black, 0);
          gl.uniform2i(this.params.cxy, this.cx, this.cy);

          gl.bindBuffer(gl.ARRAY_BUFFER, this.vbuffer);
          gl.enableVertexAttribArray(this.params.position);
          gl.vertexAttribPointer(this.params.position, 2, gl.FLOAT, false, 0, 0);

          gl.bindBuffer(gl.ARRAY_BUFFER, this.tbuffer);
          gl.enableVertexAttribArray(this.params.texCoordinate);
          gl.vertexAttribPointer(this.params.texCoordinate, 2, gl.FLOAT, false, 0, 0);

          gl.uniform2f(this.params.uSize, canvas.width, canvas.height);
          gl.uniform1f(this.params.uSetScalar0, this.mScalars[0]);
          gl.uniform1f(this.params.uSetScalar1, this.mScalars[1]);
          // gl.uniform1f(this.params.uSetScalar2, this.mScalars[2]);
          // gl.uniform1f(this.params.uSetScalar3, this.mScalars[3]);

          gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.ibuffer);
          // gl.drawElements(gl.TRIANGLES, 6 * this.pwidth * this.pheight, gl.UNSIGNED_SHORT, 0);
          // gl.drawElements(gl.TRIANGLES, 6, gl.UNSIGNED_SHORT, 2 * 6 * (this.cx + this.halfpwidth + (this.cy + this.halfpheight) * this.pwidth));
          gl.uniform1f(this.params.stage, this.mScalars[2]);
          for (let j = 0; j < this.pheight; j++) {
            for (let i = 0; i < this.pwidth; i++) {
              const offset = i + j * this.pwidth;
              const asp = this.pieces[offset].atstartpoint;
              const stage = (asp <= this.mScalars[3] ? 0.0 : ((this.mScalars[3] - asp) / (1.0 - asp)));
              // console.log(`stage ${stage} ${asp} ${this.mScalars[3]}`);
              gl.uniform1f(this.params.stage, stage);
              // gl.uniform1f(this.params.stage, (asp > this.mScalars[3]) ? ((this.mScalars[2] - asp) * (asp - this.mScalars[3]) / (1.0 - asp)) : 0.0);
              gl.uniform1f(this.params.randomrotation, this.pieces[offset].randomrotation);
              gl.uniform2f(this.params.middlepoint, this.pieces[offset].middlepoint[0], this.pieces[offset].middlepoint[1]);
              gl.uniform2f(this.params.randompoint, this.pieces[offset].randompoint[0], this.pieces[offset].randompoint[1]);
              gl.uniform2i(this.params.cxy, -this.halfpwidth + this.pieces[offset].i, -this.halfpheight + this.pieces[offset].j);
              gl.drawElements(gl.TRIANGLES, 6, gl.UNSIGNED_SHORT, 2 * 6 * offset);
            }
          }
        }

        startFilter(gl) {
          this.params.position = gl.getAttribLocation(this.program, 'position');
          this.params.texCoordinate = gl.getAttribLocation(this.program, 'texCoordinate');
          this.params.uTexture0 = gl.getUniformLocation(this.program, 'uTexture0');
          this.params.uSize = gl.getUniformLocation(this.program, 'uSize');
          this.params.uSetScalar0 = gl.getUniformLocation(this.program, 'uSetScalar0');
          this.params.uSetScalar1 = gl.getUniformLocation(this.program, 'uSetScalar1');
          // this.params.uSetScalar2 = gl.getUniformLocation(this.program, 'uSetScalar2');
          // this.params.uSetScalar3 = gl.getUniformLocation(this.program, 'uSetScalar3');
          this.params.black = gl.getUniformLocation(this.program, 'black');
          this.params.cxy = gl.getUniformLocation(this.program, 'cxy');
          this.params.stage = gl.getUniformLocation(this.program, 'stage');
          this.params.middlepoint = gl.getUniformLocation(this.program, 'middlepoint');
          this.params.randompoint = gl.getUniformLocation(this.program, 'randompoint');
          this.params.randomrotation = gl.getUniformLocation(this.program, 'randomrotation');
          this.vbuffer = gl.createBuffer();
          this.ibuffer = gl.createBuffer();
          this.tbuffer = gl.createBuffer();
        }

        setT(t) {
          this.setScalar(2, t);
          this.setScalar(3, t);
        }
    };
    return Puzzle;
  }
// define globally if it doesn't already exist
  if (typeof (Puzzle) === 'undefined') {
    const mwindow = window;
    mwindow.Puzzle = defineLibrary();
  }
}(window)); // eslint-disable-line no-undef
