(function capsule(window) {
  function defineLibrary() {
    const SceneLib = {};
    const Vec3 = VemaLib.Vec3;
    const Matrix4 = VemaLib.Matrix4;
    class Scene {
      constructor(gl, canvas) {
        this.t = 0;
        this.numOfSwarms = 1;
        this.boids = [];
        this.stage = 0;
        this.swarmCounter = 0;
        this.cb = null;
        this.textboxoffset = -1.0;
        this.route = new CamRouteLib.CamRoute(20, 200);
        console.log(`Ground height ${GroundLib.ground(0, 0)}`);
      }

      init(gl, canvas, textElem, callback) {
        if (this.cb === null) {
          this.cb = callback;
        }
        if (this.stage === 0) {
          this.stage += 1;
          textElem.innerHTML += '<p>Creating Tree</p>';
          setTimeout(() => this.init(gl, canvas, textElem), 10);
          return;
        }
        if (this.stage === 1) {
          this.tree = new PlantsLib.Spruce(gl, canvas, [0, GroundLib.ground(0, 0), 0]);
          this.stage += 1;
          textElem.innerHTML += '<p>Creating Butterfly Swarms</p>';
          setTimeout(() => this.init(gl, canvas, textElem), 10);
          return;
        }
        if (this.swarmCounter < this.numOfSwarms) {
          this.swarmCounter++;
          const swarm = new SwarmLib.Swarm(gl, canvas, DataLib.data, [0, GroundLib.ground(0, 0), 0]);
          swarm.init(gl, canvas, textElem, () => this.init(gl, canvas, textElem));
          this.boids.push(swarm);
          return;
        }
        if (this.stage === 2) {
          this.stage += 1;
          setTimeout(() => this.init(gl, canvas, textElem), 10);
          return;
        }
        if (this.stage === 3) {
          this.stage += 1;
          textElem.innerHTML += '<p>Creating Ground Patch</p>';
          this.ground = new GroundLib.Ground(gl, canvas, 200.0, 24);
          setTimeout(() => this.init(gl, canvas, textElem), 10);
          return;
        }
        if (this.stage === 4) {
          this.stage += 1;
          textElem.innerHTML += '<p>Creating Grass</p>';
          this.grass = new GrassLib.Grass(gl, canvas);
          setTimeout(() => this.init(gl, canvas, textElem), 10);
          return;
        }
        if (this.stage === 5) {
          this.stage += 1;
          textElem.innerHTML += '<p>Creating Info Box</p>';
          this.textbox = new TextBoxLib.TextBox(gl, canvas.width / 5, canvas.height / 10, 'This should not be visible');
          setTimeout(() => this.init(gl, canvas, textElem), 10);
          return;
        }
        if (this.stage === 6) {
          this.stage += 1;
          textElem.innerHTML += '<p>Creating Sky</p>';
          this.sky = new SkyLib.Sky(gl);
          setTimeout(() => this.init(gl, canvas, textElem), 10);
          return;
        }
        this.cb();
      }

      tick(gl, canvas, t) {
        this.t = t;
        this.tree.setT(t);
        if (t < 20) {
          this.textbox.renderText(gl, 'Tree has over six million vertices');
          if (t < 5) {
            this.textboxoffset = 1.0 - (t / 5);
          } else if (t > 15) {
            this.textboxoffset = 1.0 - (1.0 - ((t - 15) / 5));
          } else {
            this.textboxoffset = 0;
          }
        } else if (t > 25 && t < 45) {
          this.textbox.renderText(gl, `There are ${this.boids[0].swarm.N} butterflies in scene`);
          if (t < 30) {
            this.textboxoffset = 1.0 - ((t - 25) / 5);
          } else if (t > 40) {
            this.textboxoffset = 1.0 - (1.0 - ((t - 40) / 5));
          } else {
            this.textboxoffset = 0;
          }
        } else {
          this.textboxoffset = -1.0;
        }
        for (let i = 0; i < this.numOfSwarms; i++) { this.boids[i].tick(gl, canvas, t); }
      }

      runPhysics(t) {
        for (let i = 0; i < this.numOfSwarms; i++) { this.boids[i].runPhysics(t); }
      }

      setColorAdd(c) {
        for (let i = 0; i < this.numOfSwarms; i++) { this.boids[i].setColorAdd(c); }
      }

      setSpeed(s) {
        for (let i = 0; i < this.numOfSwarms; i++) { this.boids[i].setSpeed(s); }
      }

      draw(gl, canvas) {
        const aspectRatio = canvas.width / canvas.height;
        const fov = 45;
        const viewdistance = 0.01;
        this.perspectiveMatrix = new Matrix4();
        this.cameraMatrix = new Matrix4();
        this.model = new Matrix4();
        this.view = new Matrix4();
        this.modelPos = new Vec3(0.0, 0.0, 0.0);


        if (this.cameraStyle === 'followBF') {
          const possi = this.boids[0].swarm.possi;
          const velli = this.boids[0].swarm.velli;
          const velNorm = velli.normalized();
          const horiz = Math.atan2(velNorm.f[2], velNorm.f[0]);
          this.cameraNormal = new Vec3(-Math.cos(horiz), 0, -Math.sin(horiz));
          this.cameraPos = possi.add(this.cameraNormal.multiple(0.5));
          this.cameraUp = new Vec3(0.0, 1.0, 0.0);
        } else if (this.cameraStyle === 'followPath') {
          const p = this.route.getPoint(this.t);
          const po = [p[0], p[1], p[2]];
          // const d = this.route.getDirection(this.t - 20);
          const d = [0.0 - po[0], 2.5 + GroundLib.ground(0, 0) - po[1], 0.0 - po[2]];
          // this.cameraNormal = new Vec3(0.0, 0.0, 1.0);
          this.cameraNormal = new Vec3(-d[0], -d[1], -d[2]).normalized();
          this.cameraPos = new Vec3(po[0], po[1], po[2]);
          this.cameraUp = new Vec3(0.0, 1.0, 0.0);
          const right = this.cameraNormal.cross(this.cameraUp).normalized();
          this.cameraUp = right.cross(this.cameraNormal);
        } else {
          this.cameraNormal = new Vec3(0.0, 0.0, 1.0);
          this.cameraPos = new Vec3(0.0, 2.0 + GroundLib.ground(0, 0), 8.0);
          this.cameraUp = new Vec3(0.0, 1.0, 0.0);
        }
        this.cameraMatrix.makeLookAt(this.cameraNormal, this.cameraUp, this.cameraNormal.cross(this.cameraUp).normalized(), this.cameraPos);
        this.perspectiveMatrix.makePerspective(fov, aspectRatio, viewdistance, 2000.0);

        gl.clearColor(0.9, 0.9, 0.9, 1.0);
        // Enable depth testing
        gl.enable(gl.DEPTH_TEST);
        // Near things obscure far things
        gl.depthFunc(gl.LEQUAL);
        // Clear the color as well as the depth buffer.
        gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT); // eslint-disable-line no-bitwise
        gl.viewport(0, 0, canvas.width, canvas.height);
        this.sky.draw(gl, canvas, this.perspectiveMatrix, this.view, this.cameraMatrix, this.model, this.cameraPos.f[1]);
        this.ground.draw(gl, canvas, this.perspectiveMatrix, this.view, this.cameraMatrix, this.model);
        this.grass.draw(gl, canvas, this.perspectiveMatrix, this.cameraMatrix, this.view, this.model);
        this.tree.draw(gl, canvas, this.perspectiveMatrix, this.view, this.cameraMatrix, this.model);
        for (let i = 0; i < this.numOfSwarms; i++) { this.boids[i].draw(gl, canvas, this.perspectiveMatrix, this.view, this.cameraMatrix, this.model); }
        this.textbox.draw(gl, canvas.width / 5, canvas.height / 10, 0, 0, this.textboxoffset);
      }
    }

    SceneLib.Scene = Scene;

    return SceneLib;
  }
// define globally if it doesn't already exist
  if (typeof (SceneLib) === 'undefined') {
    const mwindow = window;
    mwindow.SceneLib = defineLibrary();
  }
}(window)); // eslint-disable-line no-undef
