(function capsule(window) {
  function defineLibrary() {
    const Vec3 = VemaLib.Vec3;
    const SimplifiedBoidsLib = {};

    class Boids {
      constructor(N) {
        this.width = 5.0;
        this.height = 8;
        this.offsetx = 0.0;
        this.offsety = 1.5;
        this.offsetz = 0.0;
        this.maxHeight = 100;
        this.minHeight = -10;
        this.center = new Vec3(this.offsetx, this.offsety, this.offsetz);
        this.maxN = N;
        this.N = 0;
        this.vel = [];
        this.pos = [];
        this.color = [];
        this.size = [];

    // min dist of approach
        this.minDist = 0.4;
    // max magnitude of velocities calculated by "rules"
        this.maxRuleVel = 0.0005;
    // max maginitude of final velocity
        this.maxVel = 0.02;
      }


      addBoids(x, y, z) {
        if (this.N < this.maxN - 1) {
          const angle = 2.0 * Math.PI * Math.random();
          const anglesup = 0.2 * Math.PI * Math.random();
          this.vel.push(new Vec3(Math.sin(angle), Math.sin(anglesup), Math.cos(angle)));
          this.pos.push(new Vec3(x, y, z));
          const r = Math.random() * 0.1;
          this.color.push(new Vec3(0.1 + r + Math.random() * 0.1, 0.1 + r + Math.random() * 0.1, 0.1 + r + Math.random() * 0.1));
          this.size.push(0.15 + 0.05 * Math.random());
          this.N += 1;
        }
      }


      limitVec(vec, maxVal) {
    // """limit magnitide of 2D vector"""
        const mag = vec.length;
        if (mag > maxVal) {
          vec.f[0] = vec.f[0] * maxVal / mag;
          vec.f[1] = vec.f[1] * maxVal / mag;
          vec.f[2] = vec.f[2] * maxVal / mag;
        }
      }

      limit(X, maxVal) {
    // """limit magnitide of 2D vectors in array X to maxValue"""
        for (let i = 0; i < this.N; i++) {
          this.limitVec(X[i], maxVal);
        }
      }

      calculateDistanceMatrix(pos) {
        const m = [];
        for (let i = 0; i < this.N; i++) {
          const r = [];
          for (let j = 0; j < this.N; j++) {
            r.push(0.0);
          }
          m.push(r);
        }
        for (let i = 0; i < this.N; i++) {
          for (let j = i + 1; j < this.N; j++) {
            const l = (pos[i].subtract(pos[j])).length;
            m[i][j] = l;
            m[j][i] = l;
          }
        }
        return m;
      }

      calculateLessThanMatrix(ms, value, multi) {
    // console.log(ms);
        const m = [];
        for (let i = 0; i < this.N; i++) {
          const r = [];
          for (let j = 0; j < this.N; j++) {
            const v = ms[i][j];
            if (v < value + (value * this.size[i] * this.size[j] - value) * multi) {
              r.push(v);
            } else {
              r.push(0.0);
            }
          }
          m.push(r);
        }
        return m;
      }

      tick() {
        this.distMatrix = this.calculateDistanceMatrix(this.pos);
        const velAdd = this.applyRules();
        for (let i = 0; i < this.N; i++) {
          this.vel[i].addToThis(velAdd[i]);
        }

        for (let i = 0; i < this.N; i++) {
          this.limitVec(this.vel[i], this.maxVel);
        }
        for (let i = 0; i < this.N; i++) { this.pos[i].addToThis(this.vel[i]); }
      }

      applyRules() {
        const vel = [];
        for (let i = 0; i < this.N; i++) { vel.push(new Vec3(0.0, 0.0, 0.0)); }
        const D = this.calculateLessThanMatrix(this.distMatrix, this.minDist, 0.0);
        for (let i = 0; i < this.N; i++) {
          let sum = 0.0;
          const msum = new Vec3(0.0, 0.0, 0.0);
          for (let j = 0; j < this.N; j++) {
            sum += D[i][j];
            msum.addToThis(this.pos[j].multiple(D[i][j]));
          }
          const vel1 = this.pos[i].multiple(sum).subtract(msum);
          this.limitVec(vel1, this.maxRuleVel);
          vel[i].addToThis(vel1);
        }
    // # different distance threshold
        const D2 = this.calculateLessThanMatrix(this.distMatrix, this.minDist * 4.0, 0.0);

    // # apply rule #2 - Alignment
        for (let i = 0; i < this.N; i++) {
          const msum = new Vec3(0.0, 0.0, 0.0);
          for (let j = 0; j < this.N; j++) {
            msum.addToThis(this.vel[j].multiple(D2[i][j]));
          }
          this.limitVec(msum, this.maxRuleVel);
          vel[i].addToThis(msum);
        }

    // # apply rule #3 - Cohesion
        for (let i = 0; i < this.N; i++) {
          const msum = new Vec3(0.0, 0.0, 0.0);
          for (let j = 0; j < this.N; j++) {
            msum.addToThis(this.pos[j].multiple(D2[i][j]));
          }
          const vel3 = msum.subtract(this.pos[i]);
          this.limitVec(vel3, this.maxRuleVel);
          vel[i].addToThis(vel3);
        }

    // # apply rul #4 stay close to center
        const maxRadius = this.width / 4;
        for (let i = 0; i < this.N; i++) {
          const n = this.center.subtract(this.pos[i]).length;
          if (n > maxRadius) {
            const nv = this.pos[i].subtract(this.center).multiple(1 / n);
            const nv2 = nv.multiple(2.0);
            const opposite = (nv.subtract(nv2)).multiple((n - maxRadius) / maxRadius * this.maxRuleVel);
            this.limitVec(opposite, this.maxRuleVel * 8.0);
            vel[i].addToThis(opposite);
          }
        }

    // # rule avoid trees
        for (let i = 0; i < this.N; i++) {
          const vel4 = new Vec3(0.0, 0.0, 0.0);
          const direction = new Vec3(0.0, Math.min(4.25, this.pos[i].f[1]), 0.0).subtract(this.pos[i]);
          const n = direction.lengthSquared;
          const dist = (4.25 - this.pos[i].f[1]);
          if (n < Math.max(0.0, dist * dist)) {
            vel4.addToThis(direction.multiple(1.0 / n * 0.2 * (-n)));
          }
          this.limitVec(vel4, this.maxRuleVel * 2.0);
          vel[i].addToThis(vel4);
        }

    // rule # 6 make random moves
        for (let i = 0; i < this.N; i++) {
          let angles = Math.PI * (Math.random() * 1.5 - 0.75);
          const x = this.vel[i].f[0];
          const y = this.vel[i].f[1];
          const z = this.vel[i].f[2];
          angles += Math.atan2(z, x);
          const horiz = Math.sqrt(x * x + z * z);
          const anglesup = Math.PI * (0.02 * Math.random() - 0.01) + Math.atan2(y, horiz);
          const vel6 = new Vec3(Math.cos(angles), Math.sin(anglesup), Math.sin(angles));
          this.limitVec(vel6, this.maxRuleVel / 3.0);
          vel[i].addToThis(vel6);
        }

        return vel;
      }
    }

    SimplifiedBoidsLib.Boids = Boids;

    return SimplifiedBoidsLib;
  }
// define globally if it doesn't already exist
  if (typeof (SimplifiedBoidsLib) === 'undefined') {
    const mwindow = window;
    mwindow.SimplifiedBoidsLib = defineLibrary();
  }
}(window)); // eslint-disable-line no-undef
